#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Intermediate Synchronic Document (ISD)'''

from __future__ import annotations

import inspect
import typing
import numbers
import re
import sys
import os
import multiprocessing
from dataclasses import dataclass
from fractions import Fraction

import ttconv.model as model
import ttconv.style_properties as styles
from ttconv.config import ModuleConfiguration
from ttconv.utils import DisjointIntervals


class SignificantTimes:
  """Information on the temporal offsets at which a ContentDocument changes.
  The class emulates the behavior of a list containing temporal offsets where the document changes,
  in increasing order
  """
  def __init__(
    self,
    sig_times: typing.List[Fraction],
    doc_cache: typing.Optional[typing.Tuple[_SingleRegionDocumentCache]] = None
    ):

    self._sig_times = tuple(sig_times)
    self._cache = doc_cache or []

  def __getitem__(self, key):
    return self._sig_times[key]

  def __len__(self):
    return len(self._sig_times)

  def __iter__(self):
    return iter(self._sig_times)

  def cache(self) -> typing.Optional[typing.Tuple[_SingleRegionDocumentCache]]:
    """Returns document cache. This dictionary is optionally used by `ISD.from_model` to increase performance.
    """
    return self._cache

  def offsets(self) -> typing.Tuple[Fraction,...]:
    """Returns a list of temporal offsets where the document changes, in increasing order.
    """
    return self._sig_times

@dataclass(frozen=True)
class _SingleRegionDocumentCache:
  """Cache for a single region document.

  `interval_cache`: maps every element in the document to its absolute temporal interval
  `doc`: document containing a single region
  `content_intervals`: set of temporal intervals during which the document is active
  """
  interval_cache: typing.Mapping[model.ContentElement, typing.Tuple[Fraction, Fraction]]
  doc: model.ContentDocument
  content_intervals: typing.Optional[DisjointIntervals]

ISD_NO_MULTIPROC_ENV = "ISD_NO_MULTIPROC"


@dataclass
class ISDConfiguration(ModuleConfiguration):
  """ISD configuration"""
  multi_thread: bool = True

  @classmethod
  def name(cls):
    return "isd"


class ISD(model.Document):
  '''Represents an Intermediate Synchronic Document, as described in TTML2, i.e. a snapshot
  of a `model.ContentDocument` taken at a given point in time

  # Constants

  `DEFAULT_REGION_ID`: `id` value of the default region
  '''

  class Region(model.Region):
    '''Represents an ISD Region, which, in contrast to a document region, can contain
    children element.
    '''

    def push_child(self, child):
      if not isinstance(child, model.Body):
        raise TypeError("Children of ISD regions must be body instances")

      if self.has_children():
        raise ValueError("ISD regions must contain at most one body instance")

      model.ContentElement.push_child(self, child)

    def is_presented(self) -> bool:
      '''Returns true if the region is presented, as specified in the IMSC specification'''
      bg_opacity: numbers.Number = self.get_style(styles.StyleProperties.Opacity)
      if bg_opacity == 0:
        return False

      bg_display: styles.DisplayType = self.get_style(styles.StyleProperties.Display)
      if bg_display is styles.DisplayType.none:
        return False

      bg_visibility: styles.VisibilityType = self.get_style(styles.StyleProperties.Visibility)
      if bg_visibility is styles.VisibilityType.hidden:
        return False

      if self.has_children():
        return True

      bg_active = self.get_style(styles.StyleProperties.ShowBackground)
      if bg_active is styles.ShowBackgroundType.whenActive:
        return False

      bg_color: styles.ColorType = self.get_style(styles.StyleProperties.BackgroundColor)
      if bg_color is None:
        return False
      if bg_color.ident is not styles.ColorType.Colorimetry.RGBA8:
        raise RuntimeError(f"Unsupported colorimetry system: {bg_color.ident}")
      # test if the background color is transparent
      if bg_color.components[3] == 0:
        return False

      return True

  DEFAULT_REGION_ID = "default_region"

  def __init__(self, doc: typing.Optional[model.ContentDocument]):
    super().__init__()

    self._regions: typing.Mapping[str, ISD.Region] = {}

    if doc is not None:
      self.set_active_area(doc.get_active_area())
      self.set_cell_resolution(doc.get_cell_resolution())
      self.set_display_aspect_ratio(doc.get_display_aspect_ratio())
      self.set_lang(doc.get_lang())
      self.set_px_resolution(doc.get_px_resolution())

  # regions

  def put_region(self, region: ISD.Region):
    '''Adds a region to the ISD, replacing any existing region with the same `id`.'''
    if not isinstance(region, ISD.Region):
      raise TypeError("Argument must be an instance of Region")

    if region.get_doc() != self:
      raise ValueError("Region does not belong to this document")

    self._regions[region.get_id()] = region

  def remove_region(self, region_id: str):
    '''Removes the region with `id == region_id` from the ISD.'''

    del self._regions[region_id]

  def get_region(self, region_id) -> typing.Optional[ISD.Region]:
    '''Returns the region with `id == region_id` or None, if none exists.'''
    return self._regions.get(region_id)

  def iter_regions(self) -> typing.Iterator[ISD.Region]:
    '''Returns an iterator over regions.'''
    return iter(self._regions.values())
  
  def __iter__(self) -> typing.Iterator[ISD.Region]:
    '''Returns an iterator over the children of the element.'''
    return iter(self.iter_regions())

  def __len__(self) -> int:
    '''Returns the number of regions of the ISD.'''
    return len(self._regions)

  @staticmethod
  def _make_absolute(
      begin_offset: typing.Optional[Fraction],
      end_offset: typing.Optional[Fraction],
      parent_begin: typing.Optional[Fraction],
      parent_end: typing.Optional[Fraction]
    ) -> typing.Tuple[Fraction, Fraction]:
    
    begin_time = (parent_begin if parent_begin is not None else Fraction(0)) + \
                  (begin_offset if begin_offset is not None else Fraction(0))

    end_time = (parent_begin if parent_begin is not None else Fraction(0)) + \
              end_offset if end_offset is not None else None

    if end_time is None:

      end_time = parent_end

    elif parent_end is not None:

      end_time = min(end_time, parent_end)

    return (begin_time, end_time)

  def _region_always_has_background(region: typing.Type[model.Region]) -> bool:

    bg_opacity: numbers.Number = region.get_style(styles.StyleProperties.Opacity)
    if bg_opacity is None:
      bg_opacity = region.get_doc().get_initial_value(styles.StyleProperties.Opacity)
    if bg_opacity == 0:
      return False

    bg_display: styles.DisplayType = region.get_style(styles.StyleProperties.Display)
    if bg_display is None:
      bg_display = region.get_doc().get_initial_value(styles.StyleProperties.Display)
    if bg_display is styles.DisplayType.none:
      return False

    bg_visibility: styles.VisibilityType = region.get_style(styles.StyleProperties.Visibility)
    if bg_visibility is None:
      bg_visibility = region.get_doc().get_initial_value(styles.StyleProperties.Visibility)
    if bg_visibility is styles.VisibilityType.hidden:
      return False

    bg_active = region.get_style(styles.StyleProperties.ShowBackground)
    if bg_active is None:
      bg_active = region.get_doc().get_initial_value(styles.StyleProperties.ShowBackground)
    if bg_active is styles.ShowBackgroundType.whenActive:
      return False

    bg_color: styles.ColorType = region.get_style(styles.StyleProperties.BackgroundColor)
    if bg_color is None:
      bg_color = region.get_doc().get_initial_value(styles.StyleProperties.BackgroundColor)
    if bg_color is not None:
      if bg_color.ident is not styles.ColorType.Colorimetry.RGBA8:
        raise RuntimeError(f"Unsupported colorimetry system: {bg_color.ident}")

      if bg_color.components[3] == 0:
        return False
    else:
      # no background color specified, so transparent
      return False

    return True

  @staticmethod
  def significant_times(doc: model.ContentDocument) -> SignificantTimes:
    '''Returns a list of the temporal offsets at which the document `doc` changes, sorted in
    increasing order'''

    def compute_sig_times(
      interval_cache,
      content_intervals: DisjointIntervals,
      s_times,
      element: model.ContentElement,
      parent_begin: Fraction,
      parent_end: typing.Optional[Fraction]
      ):

      # compute element interval

      begin_time, end_time = ISD._make_absolute(element.get_begin(), element.get_end(), parent_begin, parent_end)

      interval_cache[element] = (begin_time, end_time)

      if end_time is not None and end_time <= begin_time:
        return

      if isinstance(element, (model.Br, model.Span)) or \
          isinstance(element, (model.Region)) and ISD._region_always_has_background(element):
        content_intervals.add(begin_time, end_time)

      # add significant times for the element

      s_times.add(begin_time)

      if end_time is not None:
        s_times.add(end_time)

      # add signficant times for any animation step 

      for anim_step in element.iter_animation_steps():
        anim_begin_time, anim_end_time = ISD._make_absolute(anim_step.begin, anim_step.end, parent_begin, parent_end)

        if anim_end_time is not None and anim_end_time <= anim_begin_time:
          continue

        s_times.add(anim_begin_time)

        if anim_end_time is not None:
          s_times.add(anim_end_time)

      # add signficant times for the children of the element 

      for child_element in iter(element):
        compute_sig_times(interval_cache, content_intervals, s_times, child_element, begin_time, end_time)

    single_regions_docs = []

    s_times = set()

    doc_regions = list(doc.iter_regions())

    if len(doc_regions) > 1:
      for region in doc.iter_regions():
        single_regions_docs.append(_clone_doc_with_one_region(doc, region.get_id()))
    else:
      single_regions_docs.append(doc)


    cache = []

    for cached_doc in single_regions_docs:

      interval_cache = {}

      content_intervals = DisjointIntervals()

      # add significant times for regions

      for region in cached_doc.iter_regions():
        compute_sig_times(interval_cache, content_intervals, s_times, region, 0, None)

      # add significant times for body and its descendents

      if cached_doc.get_body() is not None:
        compute_sig_times(interval_cache, content_intervals, s_times, cached_doc.get_body(), 0, None)

      cache.append(_SingleRegionDocumentCache(
        interval_cache,
        cached_doc,
        content_intervals
        ))

    return SignificantTimes(sorted(s_times), tuple(cache))

  @staticmethod
  def from_model(
    doc: model.ContentDocument,
    offset: Fraction,
    sig_times: typing.Optional[SignificantTimes] = None) -> typing.Optional[ISD]:
    '''Creates an ISD from a snapshot of a ContentDocument `doc` at a given time offset `offset`.
    A `SignificantTimes` instance generated from `doc` can be provided to speed-up the generation process.
    '''
    isd = ISD(doc)

    cache = (_SingleRegionDocumentCache({}, doc, None),) if sig_times is None else sig_times.cache()

    for cached_doc in cache:
      if cached_doc.content_intervals is not None:
        if not cached_doc.content_intervals.contains(offset):
          continue

      regions = tuple(cached_doc.doc.iter_regions())

      activity_cache = {}

      if regions:
        for region in regions:
          isd_region = ISD._process_element(cached_doc.interval_cache, activity_cache, isd, offset, region, None, None, None, None, region)
          if isd_region is not None:
            isd.put_region(isd_region)
      else:
        default_region = model.Region(ISD.DEFAULT_REGION_ID, doc)
        isd_region = ISD._process_element(cached_doc.interval_cache, activity_cache, isd, offset, None, None, None, None, None, default_region)
        if isd_region is not None:
          isd.put_region(isd_region)

    return isd

  # TODO: remove is_multithreaded
  @staticmethod
  def generate_isd_sequence(
    doc: model.ContentDocument,
    progress_callback=lambda _: None,
    is_multithreaded: bool = True       # pylint: disable=unused-argument
    ) -> typing.List[typing.Tuple[Fraction, ISD]]:
    """ Returns a list of duples, each consisting of a significant time in the ContentDocument `doc`
    and the corresponding `ISD` instance. The duples are sorted in order of increasing significant time.
    The `is_multithreaded` flag is not used and is kept for backwards compatibility only.
    """
  
    sig_times = ISD.significant_times(doc)

    progress_callback(0.1)

    # Compute ISDs

    isds = []
    for i, isd in enumerate(map(_generate_isd, [(doc, offset, sig_times) for offset in sig_times])):
      isds.append(isd)
      progress_callback(0.1 + 0.9 * (i + 1) / len(sig_times))

    return list(zip(sig_times, isds))


  _ORDERED_STYLE_PROPS = (
    styles.StyleProperties.FontSize,
    styles.StyleProperties.Extent,
    styles.StyleProperties.Origin,
    styles.StyleProperties.Position,
    styles.StyleProperties.LineHeight,
    styles.StyleProperties.LinePadding,
    styles.StyleProperties.RubyReserve,
    styles.StyleProperties.TextOutline,
    styles.StyleProperties.TextShadow,
    styles.StyleProperties.TextEmphasis,
    styles.StyleProperties.Padding
  )

  @staticmethod
  def _compute_styles(
      styles_to_be_computed: typing.Set[typing.Type[model.StyleProperty]],
      isd_parent: typing.Optional[model.ContentElement],
      isd_element: model.ContentElement
    ):
    '''Compute style property values. This needs to be done in a specific order since the computed
    value of some style properties depend on the computed values of others'''

    for style_prop in ISD._ORDERED_STYLE_PROPS:
      if style_prop in styles_to_be_computed:
        StyleProcessors.BY_STYLE_PROP[style_prop].compute(isd_parent, isd_element)

  @staticmethod
  def _process_element(
      interval_cache,
      activity_cache,
      isd: ISD,
      absolute_offset: Fraction,
      selected_region: model.Region,
      inherited_region: typing.Optional[model.Region],
      parent: typing.Optional[model.ContentElement],
      parent_computed_begin: typing.Optional[Fraction],
      parent_computed_end: typing.Optional[Fraction],
      element: model.ContentElement
  ) -> typing.Optional[model.ContentElement]:
    # pylint: disable=too-many-arguments,too-many-locals,too-many-branches

    # first check the activity cache and return immediate if the element is not active

    is_active = activity_cache.get(element)

    if is_active is False:

      return None

    # compute the temporal extent of the element, hopefully from the cache

    element_interval = interval_cache.get(element)

    if element_interval is None:
      element_interval = ISD._make_absolute(
        element.get_begin(),
        element.get_end(),
        parent_computed_begin,
        parent_computed_end
      )
      interval_cache[element] = element_interval

    begin_time, end_time = element_interval

    if end_time is not None and end_time <= begin_time:
      return None

    # update the activity cache if the element was not present
      
    if is_active is None:

      if (
        (begin_time is not None and begin_time > absolute_offset) or
        (end_time is not None and end_time <= absolute_offset)
      ) :
        activity_cache[element] = False
        return None

      activity_cache[element] = True

    # associated region is that associated with the element, or inherited otherwise

    associated_region = element.get_region() if element.get_region() is not None else inherited_region

    # prune the element if either:
    # * the element has children and the associated region is neither the default nor the root region
    # * the element has no children and the associated region is not the root region

    if (
        not isinstance(element, model.Region) and
        associated_region is not selected_region and
        (not element.has_children() or associated_region is not None)
      ):
      return None

    # create an ISD element

    doc = element.get_doc()

    if isinstance(element, model.Region):
      isd_element = ISD.Region(element.get_id(), isd)
    else:
      isd_element = element.__class__(isd)
      isd_element.set_id(element.get_id())

    if not isinstance(element, (model.Br, model.Text)): 
      isd_element.set_lang(element.get_lang())
      isd_element.set_space(element.get_space())

    # keep track of specified style properties

    styles_to_be_computed: typing.Set[typing.Type[model.StyleProperty]] = set()

    # copy text nodes

    if isinstance(element, model.Text):
      isd_element.set_text(element.get_text())

    # apply animation

    for anim_step in element.iter_animation_steps():

      anim_begin_time, anim_end_time = ISD._make_absolute(
        anim_step.begin,
        anim_step.end,
        begin_time,
        end_time
      )

      if anim_begin_time is not None and anim_begin_time > absolute_offset:
        continue

      if anim_end_time is not None and anim_end_time <= absolute_offset:
        continue

      if anim_end_time is not None and anim_end_time <= begin_time:
        continue

      styles_to_be_computed.add(anim_step.style_property)
      isd_element.set_style(anim_step.style_property, anim_step.value)

    # copy specified styles

    for spec_style_prop in element.iter_styles():

      if isd_element.has_style(spec_style_prop):
        # skip if the style has already been set
        continue

      styles_to_be_computed.add(spec_style_prop)
      isd_element.set_style(spec_style_prop, element.get_style(spec_style_prop))

    # direction special semantics
    # https://www.w3.org/TR/ttml2/#style-attribute-direction-special-semantics

    if isinstance(element, model.Region) and \
        (not element.has_style(styles.StyleProperties.Direction)) and \
        element.get_style(styles.StyleProperties.WritingMode) in (styles.WritingModeType.lrtb, styles.WritingModeType.rltb):
      styles_to_be_computed.add(styles.StyleProperties.Direction)
      direction = styles.DirectionType.ltr if element.get_style(styles.StyleProperties.WritingMode) == styles.WritingModeType.lrtb \
                  else styles.DirectionType.rtl
      isd_element.set_style(styles.StyleProperties.Direction, direction)

    # inherited styling

    if not isinstance(element, (model.Br, model.Text, model.Region)):

      for inherited_style_prop in parent.iter_styles():

        StyleProcessors.BY_STYLE_PROP[inherited_style_prop].inherit(parent, isd_element)


    # initial value styling

    if not isinstance(element, (model.Br, model.Text)):

      for initial_style in styles.StyleProperties.ALL:

        if isd_element.has_style(initial_style):
          continue

        if doc.has_initial_value(initial_style):

          initial_value = doc.get_initial_value(initial_style)

        elif initial_style is not styles.StyleProperties.Position:

          # the initial value of the Position style property is set to Origin as part of style computation

          initial_value = initial_style.make_initial_value()

        else:

          initial_value = None

        styles_to_be_computed.add(initial_style)

        isd_element.set_style(initial_style, initial_value)

    # compute style properties

    ISD._compute_styles(styles_to_be_computed, parent, isd_element)

    # prune element is display is "none"

    if isd_element.get_style(styles.StyleProperties.Display) is styles.DisplayType.none:
      return None

    # process children of the element

    isd_element_children = []

    if isinstance(element, model.Region):

      if doc.get_body() is not None:
        isd_body_element = ISD._process_element(
          interval_cache,
          activity_cache,
          isd,
          absolute_offset,
          selected_region,
          associated_region,
          isd_element,
          None,
          None,
          doc.get_body()
        )

        if isd_body_element is not None:
          isd_element_children.append(isd_body_element)

    else:

      for child_element in element:
        isd_element_child = ISD._process_element(
              interval_cache,
              activity_cache,
              isd,
              absolute_offset,
              selected_region,
              associated_region,
              isd_element,
              begin_time,
              end_time,
              child_element
        )

        if isd_element_child is not None:
          isd_element_children.append(isd_element_child)

    if len(isd_element_children) > 0:
      isd_element.push_children(isd_element_children)

      if isinstance(isd_element, (model.P, model.Rt, model.Rtc)):
        text_node_list = []
        _construct_text_list(isd_element, text_node_list)
        _process_lwsp(text_node_list)
        _prune_empty_spans(isd_element)

    # remove styles that are not applicable

    for style_prop in list(isd_element.iter_styles()):
      if not isd_element.is_style_applicable(style_prop):
        isd_element.set_style(style_prop, None)
    
    # prune or keep the element

    if isinstance(isd_element, (model.Br, model.Text,model.Rb, model.Rbc)):
      return isd_element

    if isd_element.has_children():
      return isd_element

    if (
        isinstance(isd_element, ISD.Region) and
        isd_element.get_style(styles.StyleProperties.ShowBackground) is styles.ShowBackgroundType.always
      ):
      return isd_element

    return None

def _prune_empty_spans(element: model.ContentElement):
  children = list(element)
  for child in children:
    _prune_empty_spans(child)
    if isinstance(child, model.Text) and not child.get_text():
      element.remove_child(child)
    elif isinstance(child, model.Span) and not child:
      element.remove_child(child)


def _construct_text_list(element: model.ContentElement, text_node_list: typing.List[typing.Union[model.Text, model.Br]]):
  '''Constructs a list of all text and br elements in dfs order, excluding rt, rtc and rp elements'''
  for child in element:
    if isinstance(child, model.Br) or (isinstance(child, model.Text) and child.get_text()):
      text_node_list.append(child)
    elif not isinstance(child, (model.Rt, model.Rtc, model.Rp)):
      _construct_text_list(child, text_node_list)

def _process_lwsp(text_node_list: typing.List[typing.Union[model.Text, model.Br]]):
  '''Processes LWSP according to the space property'''

  def _is_prev_char_lwsp(prev_element):
    if isinstance(prev_element, model.Br):
      return True
      
    prev_text = prev_element.get_text()

    return len(prev_text) > 0 and prev_text[-1] in ("\t", "\r", "\n", " ")

  def _is_next_char_lwsp(next_element):
    if isinstance(next_element, model.Br):
      return True

    next_text = next_element.get_text()

    return len(next_text) > 0 and next_text[0] in ("\r", "\n")


  elist = list(text_node_list)

  # first pass: collapse spaces and remove leading LWSPs

  i = 0

  while i < len(elist):

    node = elist[i]

    if isinstance(node, model.Br) or node.parent().get_space() is model.WhiteSpaceHandling.PRESERVE:
      i += 1
      continue

    trimmed_text = re.sub(r"[\t\r\n ]+", " ", node.get_text())

    if len(trimmed_text) > 0 and trimmed_text[0] == " ":

      if i == 0 or _is_prev_char_lwsp(elist[i - 1]):

        trimmed_text = trimmed_text[1:]

    node.set_text(trimmed_text)

    if len(trimmed_text) == 0:
      del elist[i]
    else:
      i += 1

  # second pass: remove trailing LWSPs

  for i, node in enumerate(elist):

    if isinstance(node, model.Br) or node.parent().get_space() is model.WhiteSpaceHandling.PRESERVE:
      i += 1
      continue

    node_text = node.get_text()

    if node_text[-1] == " ":

      if i == (len(elist) - 1) or _is_next_char_lwsp(elist[i + 1]):

        node.set_text(node_text[:-1])

    

def _compute_length(
    source_length: styles.LengthType,
    pct_ref: typing.Optional[styles.LengthType],
    em_ref: typing.Optional[styles.LengthType],
    c_ref: typing.Optional[styles.LengthType],
    px_ref: typing.Optional[styles.LengthType]
  ):

  if source_length.units is styles.LengthType.Units.pct:
    if pct_ref is None:
      raise ValueError("Percent length computed without pct reference")

    return styles.LengthType(
      value=source_length.value * pct_ref.value / 100,
      units=pct_ref.units
    )

  if source_length.units is styles.LengthType.Units.em:
    if em_ref is None:
      raise ValueError("Em length computed without em reference")

    return styles.LengthType(
      value=source_length.value * em_ref.value,
      units=em_ref.units
    )

  if source_length.units is styles.LengthType.Units.c:
    if c_ref is None:
      raise ValueError("C length computed without c reference")

    return styles.LengthType(
      value=source_length.value * c_ref.value,
      units=c_ref.units
    )

  if source_length.units is styles.LengthType.Units.px:
    if px_ref is None:
      raise ValueError("px length computed without px reference")

    return styles.LengthType(
      value=source_length.value * px_ref.value,
      units=px_ref.units
    )

  return source_length

def _make_rh_length(value: numbers.Number) -> styles.LengthType:
  '''Creates a length expressed in `rh` units
  '''
  return styles.LengthType(
    value=value,
    units=styles.LengthType.Units.rh
  )

def _make_rw_length(value: numbers.Number) -> styles.LengthType:
  '''Creates a length expressed in `rw` units
  '''
  return styles.LengthType(
    value=value,
    units=styles.LengthType.Units.rw
  )

def _get_writing_mode(isd_parent: model.ContentElement, isd_element: model.ContentElement) -> styles.WritingModeType:

  while isd_parent is not None:
    isd_element = isd_parent
    isd_parent = isd_element.parent()

  return isd_element.get_style(styles.StyleProperties.WritingMode)

class StyleProcessor:
  '''Processes style properties during the style resolution process.
  
  Class variables:

  * `style_prop`: reference to the style property that the processor handles
  '''

  style_prop: typing.Type[styles.StyleProperty] = None

  @classmethod
  def compute(cls, parent: model.ContentElement, element: model.ContentElement):
    '''Reads the (specified) style property value and computes it
    '''

  @classmethod
  def inherit(cls, parent: model.ContentElement, element: model.ContentElement):
    '''Inherit the style property from the parent to the element
    '''
    if cls.style_prop.is_inherited and not element.has_style(cls.style_prop):
      element.set_style(cls.style_prop, parent.get_style(cls.style_prop))

class StyleProcessors:
  '''Processes style properties during the style resolution process
  
  Class variables:

  * `BY_STYLE_PROP`: maps a style property to a processor
  '''

  # pylint: disable=missing-class-docstring

  class BackgroundColor(StyleProcessor):
    style_prop = styles.StyleProperties.BackgroundColor

  class Color(StyleProcessor):
    style_prop = styles.StyleProperties.Color

  class Direction(StyleProcessor):
    style_prop = styles.StyleProperties.Direction

  class Disparity(StyleProcessor):
    style_prop = styles.StyleProperties.Disparity

  class Display(StyleProcessor):
    style_prop = styles.StyleProperties.Display

  class DisplayAlign(StyleProcessor):
    style_prop = styles.StyleProperties.DisplayAlign

  class Extent(StyleProcessor):
    style_prop = styles.StyleProperties.Extent

    @classmethod
    def compute(cls, parent: model.ContentElement, element: model.ContentElement):

      style_value: styles.ExtentType = element.get_style(cls.style_prop)

      # height

      height = _compute_length(
        style_value.height,
        _make_rh_length(100),
        element.get_style(styles.StyleProperties.FontSize),
        _make_rh_length(100 / element.get_doc().get_cell_resolution().rows),
        _make_rh_length(100 / element.get_doc().get_px_resolution().height)
      )

      # width

      width = _compute_length(
        style_value.width,
        _make_rw_length(100),
        element.get_style(styles.StyleProperties.FontSize),
        _make_rw_length(100 / element.get_doc().get_cell_resolution().columns),
        _make_rw_length(100 / element.get_doc().get_px_resolution().width)
      )

      element.set_style(
        cls.style_prop,
        styles.ExtentType(
          height=height,
          width=width
        )
      )

  class FillLineGap(StyleProcessor):
    style_prop = styles.StyleProperties.FillLineGap

  class FontFamily(StyleProcessor):
    style_prop = styles.StyleProperties.FontFamily

  class FontSize(StyleProcessor):
    style_prop = styles.StyleProperties.FontSize

    @classmethod
    def inherit(cls, parent: model.ContentElement, element: model.ContentElement):
      if element.has_style(cls.style_prop):
        return

      parent_value: styles.LengthType = parent.get_style(cls.style_prop)

      if (
          isinstance(element, model.Rtc) or
          (isinstance(element, model.Rt) and not isinstance(parent, model.Rtc))
      ):
        style_value = styles.LengthType(
          value=parent_value.value/2,
          units=parent_value.units
        )
      else:
        style_value = parent_value

      element.set_style(cls.style_prop, style_value)

    @classmethod
    def compute(cls, parent: model.ContentElement, element: model.ContentElement):

      style_value = element.get_style(cls.style_prop)
      parent_value = parent.get_style(cls.style_prop) if parent is not None else None

      pct_ref = parent_value if parent_value is not None \
         else _make_rh_length(100/element.get_doc().get_cell_resolution().rows)

      element.set_style(
        cls.style_prop,
        _compute_length(
          style_value,
          pct_ref,
          pct_ref,
          _make_rh_length(100 / element.get_doc().get_cell_resolution().rows),
          _make_rh_length(100 / element.get_doc().get_px_resolution().height)
        )
      )      

  class FontStyle(StyleProcessor):
    style_prop = styles.StyleProperties.FontStyle

  class FontVariant(StyleProcessor):
    style_prop = styles.StyleProperties.FontVariant

  class FontWeight(StyleProcessor):
    style_prop = styles.StyleProperties.FontWeight

  class LineHeight(StyleProcessor):
    style_prop = styles.StyleProperties.LineHeight

    @classmethod
    def compute(cls, parent: model.ContentElement, element: model.ContentElement):
      value = element.get_style(cls.style_prop)

      if value is styles.SpecialValues.normal:

        computed_value = value

      else:
        computed_value = _compute_length(
          value,
          element.get_style(styles.StyleProperties.FontSize),
          element.get_style(styles.StyleProperties.FontSize),
          _make_rh_length(100 / element.get_doc().get_cell_resolution().rows),
          _make_rh_length(100 / element.get_doc().get_px_resolution().height)
        )

      element.set_style(cls.style_prop, computed_value)

  class LinePadding(StyleProcessor):
    style_prop = styles.StyleProperties.LinePadding

    @classmethod
    def compute(cls, parent: model.ContentElement, element: model.ContentElement):
      element.set_style(
        cls.style_prop,
        _compute_length(
          element.get_style(cls.style_prop),
          element.get_style(styles.StyleProperties.FontSize),
          element.get_style(styles.StyleProperties.FontSize),
          _make_rh_length(100 / element.get_doc().get_cell_resolution().rows),
          _make_rh_length(100 / element.get_doc().get_px_resolution().height)
        )
      )

  class LuminanceGain(StyleProcessor):
    style_prop = styles.StyleProperties.LuminanceGain

  class MultiRowAlign(StyleProcessor):
    style_prop = styles.StyleProperties.MultiRowAlign

  class Opacity(StyleProcessor):
    style_prop = styles.StyleProperties.Opacity

  class Origin(StyleProcessor):
    style_prop = styles.StyleProperties.Origin

    @classmethod
    def compute(cls, parent: model.ContentElement, element: model.ContentElement):

      style_value: styles.CoordinateType = element.get_style(cls.style_prop)

      # height

      y = _compute_length(
        style_value.y,
        _make_rh_length(100),
        None,
        _make_rh_length(100 / element.get_doc().get_cell_resolution().rows),
        _make_rh_length(100 / element.get_doc().get_px_resolution().height)
      )

      # width

      x = _compute_length(
        style_value.x,
        _make_rw_length(100),
        None,
        _make_rw_length(100 / element.get_doc().get_cell_resolution().columns),
        _make_rw_length(100 / element.get_doc().get_px_resolution().width)
      )

      element.set_style(
        cls.style_prop,
        styles.CoordinateType(
          x=x,
          y=y
        )
      )

  class Overflow(StyleProcessor):
    style_prop = styles.StyleProperties.Overflow

  class Padding(StyleProcessor):
    style_prop = styles.StyleProperties.Padding

    @classmethod
    def compute(cls, parent: model.ContentElement, element: model.ContentElement):

      padding_value: styles.PaddingType = element.get_style(cls.style_prop)

      wm: styles.WritingModeType = element.get_style(styles.StyleProperties.WritingMode)

      extent: styles.ExtentType = element.get_style(styles.StyleProperties.Extent)

      is_vertical = wm in (styles.WritingModeType.tblr, styles.WritingModeType.tbrl)

      c_h = _make_rh_length(100 / element.get_doc().get_cell_resolution().rows)
      px_h = _make_rh_length(100 / element.get_doc().get_px_resolution().height)
      c_w = _make_rw_length(100 / element.get_doc().get_cell_resolution().columns)
      px_w = _make_rw_length(100 / element.get_doc().get_px_resolution().width)

      c_before = _compute_length(
        padding_value.before,
        extent.width if is_vertical else extent.height,
        element.get_style(styles.StyleProperties.FontSize),
        c_w if is_vertical else c_h,
        px_w if is_vertical else px_h
      )

      c_after = _compute_length(
        padding_value.after,
        extent.width if is_vertical else extent.height,
        element.get_style(styles.StyleProperties.FontSize),
        c_w if is_vertical else c_h,
        px_w if is_vertical else px_h
      )

      c_start = _compute_length(
        padding_value.start,
        extent.width if not is_vertical else extent.height,
        element.get_style(styles.StyleProperties.FontSize),
        c_w if not is_vertical else c_h,
        px_w if not is_vertical else px_h
      )

      c_end = _compute_length(
        padding_value.end,
        extent.width if not is_vertical else extent.height,
        element.get_style(styles.StyleProperties.FontSize),
        c_w if not is_vertical else c_h,
        px_w if not is_vertical else px_h
      )

      element.set_style(
        cls.style_prop,
        styles.PaddingType(c_before, c_end, c_after, c_start)
      )    

  class Position(StyleProcessor):
    style_prop = styles.StyleProperties.Position

    @classmethod
    def compute(cls, parent: model.ContentElement, element: model.ContentElement):

      position : styles.PositionType = element.get_style(styles.StyleProperties.Position)

      if position is None:

        # if no Position style property was specified, use the value of the Origin style property

        origin : styles.CoordinateType = element.get_style(styles.StyleProperties.Origin)

        element.set_style(
          styles.StyleProperties.Position,
          styles.PositionType(
              h_offset=origin.x,
              v_offset=origin.y
            )
        )

        return

      extent : styles.ExtentType = element.get_style(styles.StyleProperties.Extent)

      assert extent.height.units is styles.LengthType.Units.rh
      assert extent.width.units is styles.LengthType.Units.rw

      v_offset = _compute_length(
        position.v_offset,
        _make_rh_length(100 - extent.height.value),
        None,
        _make_rh_length(100 / element.get_doc().get_cell_resolution().rows),
        _make_rh_length(100 / element.get_doc().get_px_resolution().height)
      )
      
      if position.v_edge is styles.PositionType.VEdge.bottom:
        v_offset = styles.LengthType(
          value=100 - v_offset.value,
          units=v_offset.units
        )

      h_offset = _compute_length(
        position.h_offset,
        _make_rw_length(100 - extent.width.value),
        None,
        _make_rw_length(100 / element.get_doc().get_cell_resolution().columns),
        _make_rw_length(100 / element.get_doc().get_px_resolution().width)
      )

      if position.h_edge is styles.PositionType.HEdge.right:
        h_offset = styles.LengthType(
          value=100 - h_offset.value,
          units=h_offset.units
        )

      # if specified, the Position style property overrides the Origin style property

      element.set_style(
        styles.StyleProperties.Origin,
        styles.CoordinateType(
            x=h_offset,
            y=v_offset
          )
      )

      element.set_style(
        styles.StyleProperties.Position,
        styles.PositionType(
            h_offset=h_offset,
            v_offset=v_offset
          )
      )

  class RubyAlign(StyleProcessor):
    style_prop = styles.StyleProperties.RubyAlign

  class RubyPosition(StyleProcessor):
    style_prop = styles.StyleProperties.RubyPosition

  class RubyReserve(StyleProcessor):
    style_prop = styles.StyleProperties.RubyReserve

    @classmethod
    def compute(cls, parent: model.ContentElement, element: model.ContentElement):
      value: typing.Union[styles.SpecialValues.none, styles.RubyReserveType] = element.get_style(cls.style_prop)

      if value is styles.SpecialValues.none:

        computed_value = value

      elif value.length is not None:
        computed_value = styles.RubyReserveType(
          position=value.position,
          length=_compute_length(
            value.length,
            element.get_style(styles.StyleProperties.FontSize),
            element.get_style(styles.StyleProperties.FontSize),
            _make_rh_length(100 / element.get_doc().get_cell_resolution().rows),
            _make_rh_length(100 / element.get_doc().get_px_resolution().height)
          )
        )

      else:
        fs = element.get_style(styles.StyleProperties.FontSize)
        computed_value = styles.RubyReserveType(
          position=value.position,
          length=styles.LengthType(
            fs.value/2,
            fs.units
          )
        )
        
      element.set_style(cls.style_prop, computed_value)

  class Shear(StyleProcessor):
    style_prop = styles.StyleProperties.Shear

  class ShowBackground(StyleProcessor):
    style_prop = styles.StyleProperties.ShowBackground

  class TextAlign(StyleProcessor):
    style_prop = styles.StyleProperties.TextAlign

  class TextCombine(StyleProcessor):
    style_prop = styles.StyleProperties.TextCombine

  class TextDecoration(StyleProcessor):
    style_prop = styles.StyleProperties.TextDecoration

    @classmethod
    def inherit(cls, parent: model.ContentElement, element: model.ContentElement):
      parent_value: styles.TextDecorationType = parent.get_style(cls.style_prop)

      spec_value: styles.TextDecorationType = element.get_style(cls.style_prop)

      if spec_value is None:

        style_value = parent_value

      else:
        style_value = styles.TextDecorationType(
          underline=spec_value.underline if spec_value.underline is not None else parent_value.underline,
          line_through=spec_value.line_through if spec_value.line_through is not None else parent_value.line_through,
          overline=spec_value.overline if spec_value.overline is not None else parent_value.overline
        )

      element.set_style(cls.style_prop, style_value)

  class TextEmphasis(StyleProcessor):
    style_prop = styles.StyleProperties.TextEmphasis

    @classmethod
    def compute(cls, parent: model.ContentElement, element: model.ContentElement):
      value: styles.TextEmphasisType = element.get_style(cls.style_prop)

      if value is styles.SpecialValues.none:

        computed_value = value

      else:
        
        color = value.color if value.color is not None else element.get_style(styles.StyleProperties.Color)

        if value.style is styles.TextEmphasisType.Style.auto:

          wm: styles.WritingModeType = _get_writing_mode(parent, element)

          if wm in (styles.WritingModeType.tblr, styles.WritingModeType.tbrl):

            style = styles.TextEmphasisType.Style.filled_sesame

          else:

            style = styles.TextEmphasisType.Style.filled_circle

        else:

          style = value.style

        computed_value = styles.TextEmphasisType(
          color=color,
          position=value.position,
          style=style
        )
        
      element.set_style(cls.style_prop, computed_value)


  class TextOutline(StyleProcessor):
    style_prop = styles.StyleProperties.TextOutline

    @classmethod
    def compute(cls, parent: model.ContentElement, element: model.ContentElement):
      value: typing.Union[styles.SpecialValues.none, styles.TextOutlineType] = element.get_style(cls.style_prop)

      if value is styles.SpecialValues.none:

        computed_value = value

      else:

        computed_value = styles.TextOutlineType(
          color=value.color if value.color is not None else element.get_style(styles.StyleProperties.Color),
          thickness=_compute_length(
            value.thickness,
            element.get_style(styles.StyleProperties.FontSize),
            element.get_style(styles.StyleProperties.FontSize),
            _make_rh_length(100 / element.get_doc().get_cell_resolution().rows),
            _make_rh_length(100 / element.get_doc().get_px_resolution().height)
          )
        )
        
      element.set_style(cls.style_prop, computed_value)

  class TextShadow(StyleProcessor):
    style_prop = styles.StyleProperties.TextShadow

    @classmethod
    def compute(cls, parent: model.ContentElement, element: model.ContentElement):
      value: typing.Union[styles.SpecialValues.none, styles.TextShadowType] = element.get_style(cls.style_prop)

      if value is styles.SpecialValues.none:

        computed_value = value

      else:

        shadows = []

        for shadow in value.shadows:
          shadows.append(
            styles.TextShadowType.Shadow(
              color=shadow.color if shadow.color is not None else element.get_style(styles.StyleProperties.Color),
              x_offset=_compute_length(
                shadow.x_offset,
                element.get_style(styles.StyleProperties.FontSize),
                element.get_style(styles.StyleProperties.FontSize),
                _make_rh_length(100 / element.get_doc().get_cell_resolution().rows),
                _make_rh_length(100 / element.get_doc().get_px_resolution().height)
              ),
              y_offset=_compute_length(
                shadow.y_offset,
                element.get_style(styles.StyleProperties.FontSize),
                element.get_style(styles.StyleProperties.FontSize),
                _make_rh_length(100 / element.get_doc().get_cell_resolution().rows),
                _make_rh_length(100 / element.get_doc().get_px_resolution().height)
              ),
              blur_radius=None if shadow.blur_radius is None else _compute_length(
                shadow.blur_radius,
                element.get_style(styles.StyleProperties.FontSize),
                element.get_style(styles.StyleProperties.FontSize),
                _make_rh_length(100 / element.get_doc().get_cell_resolution().rows),
                _make_rh_length(100 / element.get_doc().get_px_resolution().height)
              )
            )
          )

        computed_value = styles.TextShadowType(tuple(shadows))
        
      element.set_style(cls.style_prop, computed_value)

  class UnicodeBidi(StyleProcessor):
    style_prop = styles.StyleProperties.UnicodeBidi

  class Visibility(StyleProcessor):
    style_prop = styles.StyleProperties.Visibility

  class WrapOption(StyleProcessor):
    style_prop = styles.StyleProperties.WrapOption

  class WritingMode(StyleProcessor):
    style_prop = styles.StyleProperties.WritingMode

  BY_STYLE_PROP = {
    processor.style_prop : processor
    for processor_name, processor in list(locals().items()) if inspect.isclass(processor) and processor.style_prop is not None
    }

# pylint: enable=missing-class-docstring

def _generate_isd(args):
  doc, offset, sig_times = args
  return ISD.from_model(doc, offset, sig_times)


def _clone_doc_with_one_region(doc: model.ContentDocument, region_id: str):

  def _copy_content_element(
    new_doc: model.ContentDocument,
    selected_region: model.Region,
    inherited_region: model.Region,
    element: model.ContentElement
  ):
    if element is None:
      return None

    # associated region is that associated with the element, or inherited otherwise

    associated_region = element.get_region() if element.get_region() is not None else inherited_region

    # prune the element if either:
    # * the element has children and the associated region is neither the default nor the root region
    # * the element has no children and the associated region is not the root region

    if (
        associated_region is not selected_region and
        (not element.has_children() or associated_region is not None)
      ):
      return None

    new_element = type(element)(new_doc)
    element.copy_to(new_element)

    region = element.get_region()

    if region is not None:
      new_element.set_region(new_doc.get_region(region.get_id()))
    
    new_children = []

    for child in element:
      new_child = _copy_content_element(new_doc, selected_region, associated_region, child)
      if new_child is not None:
        new_children.append(new_child)

    if len(new_children) > 0:
      new_element.push_children(new_children)

    return new_element

  new_doc = model.ContentDocument()

  doc.copy_to(new_doc)

  region = doc.get_region(region_id)

  new_region = model.Region(region_id, new_doc)

  region.copy_to(new_region)

  new_doc.put_region(new_region)

  new_body = _copy_content_element(new_doc, doc.get_region(region_id), None, doc.get_body())

  new_doc.set_body(new_body)

  return new_doc



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Unit tests for the SCC Control codes"""

# pylint: disable=R0201,C0115,C0116

import unittest

from ttconv.scc.codes.attribute_codes import SccAttributeCode

VALID_VALUES = [0x1020, 0x1820, 0x1021, 0x1821, 0x1022, 0x1822, 0x1023, 0x1823, 0x1024, 0x1824, 0x1025, 0x1825,
                0x1026, 0x1826, 0x1027, 0x1827, 0x1028, 0x1828, 0x1029, 0x1829, 0x102A, 0x182A, 0x102B, 0x182B,
                0x102C, 0x182C, 0x102D, 0x182D, 0x102E, 0x182E, 0x102F, 0x182F, 0x172D, 0x1F2D, 0x172E, 0x1F2E,
                0x172F, 0x1F2F]


class TestSccAttributeCode(unittest.TestCase):

  def test_scc_attribute_codes(self):
    self.assertEqual(SccAttributeCode.BWO,
                     SccAttributeCode.find(VALID_VALUES[0]))
    self.assertEqual(SccAttributeCode.BWO,
                     SccAttributeCode.find(VALID_VALUES[1]))
    self.assertEqual(SccAttributeCode.BWS,
                     SccAttributeCode.find(VALID_VALUES[2]))
    self.assertEqual(SccAttributeCode.BWS,
                     SccAttributeCode.find(VALID_VALUES[3]))
    self.assertEqual(SccAttributeCode.BGO,
                     SccAttributeCode.find(VALID_VALUES[4]))
    self.assertEqual(SccAttributeCode.BGO,
                     SccAttributeCode.find(VALID_VALUES[5]))
    self.assertEqual(SccAttributeCode.BGS,
                     SccAttributeCode.find(VALID_VALUES[6]))
    self.assertEqual(SccAttributeCode.BGS,
                     SccAttributeCode.find(VALID_VALUES[7]))
    self.assertEqual(SccAttributeCode.BBO,
                     SccAttributeCode.find(VALID_VALUES[8]))
    self.assertEqual(SccAttributeCode.BBO,
                     SccAttributeCode.find(VALID_VALUES[9]))
    self.assertEqual(SccAttributeCode.BBS,
                     SccAttributeCode.find(VALID_VALUES[10]))
    self.assertEqual(SccAttributeCode.BBS,
                     SccAttributeCode.find(VALID_VALUES[11]))
    self.assertEqual(SccAttributeCode.BCO,
                     SccAttributeCode.find(VALID_VALUES[12]))
    self.assertEqual(SccAttributeCode.BCO,
                     SccAttributeCode.find(VALID_VALUES[13]))
    self.assertEqual(SccAttributeCode.BCS,
                     SccAttributeCode.find(VALID_VALUES[14]))
    self.assertEqual(SccAttributeCode.BCS,
                     SccAttributeCode.find(VALID_VALUES[15]))
    self.assertEqual(SccAttributeCode.BRO,
                     SccAttributeCode.find(VALID_VALUES[16]))
    self.assertEqual(SccAttributeCode.BRO,
                     SccAttributeCode.find(VALID_VALUES[17]))
    self.assertEqual(SccAttributeCode.BRS,
                     SccAttributeCode.find(VALID_VALUES[18]))
    self.assertEqual(SccAttributeCode.BRS,
                     SccAttributeCode.find(VALID_VALUES[19]))
    self.assertEqual(SccAttributeCode.BYO,
                     SccAttributeCode.find(VALID_VALUES[20]))
    self.assertEqual(SccAttributeCode.BYO,
                     SccAttributeCode.find(VALID_VALUES[21]))
    self.assertEqual(SccAttributeCode.BYS,
                     SccAttributeCode.find(VALID_VALUES[22]))
    self.assertEqual(SccAttributeCode.BYS,
                     SccAttributeCode.find(VALID_VALUES[23]))
    self.assertEqual(SccAttributeCode.BMO,
                     SccAttributeCode.find(VALID_VALUES[24]))
    self.assertEqual(SccAttributeCode.BMO,
                     SccAttributeCode.find(VALID_VALUES[25]))
    self.assertEqual(SccAttributeCode.BMS,
                     SccAttributeCode.find(VALID_VALUES[26]))
    self.assertEqual(SccAttributeCode.BMS,
                     SccAttributeCode.find(VALID_VALUES[27]))
    self.assertEqual(SccAttributeCode.BAO,
                     SccAttributeCode.find(VALID_VALUES[28]))
    self.assertEqual(SccAttributeCode.BAO,
                     SccAttributeCode.find(VALID_VALUES[29]))
    self.assertEqual(SccAttributeCode.BAS,
                     SccAttributeCode.find(VALID_VALUES[30]))
    self.assertEqual(SccAttributeCode.BAS,
                     SccAttributeCode.find(VALID_VALUES[31]))
    self.assertEqual(SccAttributeCode.BT,
                     SccAttributeCode.find(VALID_VALUES[32]))
    self.assertEqual(SccAttributeCode.BT,
                     SccAttributeCode.find(VALID_VALUES[33]))
    self.assertEqual(SccAttributeCode.FA,
                     SccAttributeCode.find(VALID_VALUES[34]))
    self.assertEqual(SccAttributeCode.FA,
                     SccAttributeCode.find(VALID_VALUES[35]))
    self.assertEqual(SccAttributeCode.FAU,
                     SccAttributeCode.find(VALID_VALUES[36]))
    self.assertEqual(SccAttributeCode.FAU,
                     SccAttributeCode.find(VALID_VALUES[37]))

  def test_scc_attribute_codes_invalid(self):
    other_code_values = [code for code in range(0x0000, 0xFFFF) if code not in VALID_VALUES]

    for cc in other_code_values:
      self.assertIsNone(SccAttributeCode.find(cc))

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for the WebVTT tokenizer'''

# pylint: disable=R0201,C0115,C0116

import unittest

import ttconv.vtt.tokenizer as tokenizer

class VTTTokenizerTest(unittest.TestCase):

  def test_tokenizer_classes(self):
    cue_text = "<c.yellow.bg_blue.magenta.bg_black>This is magenta text on a black background</c>"

    t = tokenizer.CueTextTokenizer(cue_text)

    start_tag_token: tokenizer.StartTagToken = next(t)
    self.assertIsInstance(start_tag_token, tokenizer.StartTagToken)
    self.assertEqual(start_tag_token.tag, "c")
    self.assertListEqual(start_tag_token.classes, ["yellow", "bg_blue", "magenta", "bg_black"])

    string_token: tokenizer.StringToken = next(t)
    self.assertIsInstance(string_token, tokenizer.StringToken)
    self.assertEqual(string_token.value, "This is magenta text on a black background")

    end_tag_token: tokenizer.EndTagToken = next(t)
    self.assertIsInstance(end_tag_token, tokenizer.EndTagToken)
    self.assertEqual(end_tag_token.tag, "c")

    with self.assertRaises(StopIteration):
      next(t)

  def test_tokenizer_multiline_incomplete_elements(self):
    cue_text = """<c.yellow><c.bg_black >This</c></c>
<c.bg_yellow>is
<c.bg_red>row 18"""

    t = tokenizer.CueTextTokenizer(cue_text)

    start_tag_token: tokenizer.StartTagToken = next(t)
    self.assertIsInstance(start_tag_token, tokenizer.StartTagToken)
    self.assertEqual(start_tag_token.tag, "c")
    self.assertListEqual(start_tag_token.classes, ["yellow"])

    start_tag_token: tokenizer.StartTagToken = next(t)
    self.assertIsInstance(start_tag_token, tokenizer.StartTagToken)
    self.assertEqual(start_tag_token.tag, "c")
    self.assertListEqual(start_tag_token.classes, ["bg_black"])

    string_token: tokenizer.StringToken = next(t)
    self.assertIsInstance(string_token, tokenizer.StringToken)
    self.assertEqual(string_token.value, "This")

    end_tag_token: tokenizer.EndTagToken = next(t)
    self.assertIsInstance(end_tag_token, tokenizer.EndTagToken)
    self.assertEqual(end_tag_token.tag, "c")

    end_tag_token: tokenizer.EndTagToken = next(t)
    self.assertIsInstance(end_tag_token, tokenizer.EndTagToken)
    self.assertEqual(end_tag_token.tag, "c")

    string_token: tokenizer.StringToken = next(t)
    self.assertIsInstance(string_token, tokenizer.StringToken)
    self.assertEqual(string_token.value, "\n")

    start_tag_token: tokenizer.StartTagToken = next(t)
    self.assertIsInstance(start_tag_token, tokenizer.StartTagToken)
    self.assertEqual(start_tag_token.tag, "c")
    self.assertListEqual(start_tag_token.classes, ["bg_yellow"])

    string_token: tokenizer.StringToken = next(t)
    self.assertIsInstance(string_token, tokenizer.StringToken)
    self.assertEqual(string_token.value, "is\n")

    start_tag_token: tokenizer.StartTagToken = next(t)
    self.assertIsInstance(start_tag_token, tokenizer.StartTagToken)
    self.assertEqual(start_tag_token.tag, "c")
    self.assertListEqual(start_tag_token.classes, ["bg_red"])

    string_token: tokenizer.StringToken = next(t)
    self.assertIsInstance(string_token, tokenizer.StringToken)
    self.assertEqual(string_token.value, "row 18")

    with self.assertRaises(StopIteration):
      next(t)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SCC Preamble Address Codes"""

from __future__ import annotations

from typing import Optional

from ttconv.scc.codes import SCC_COLOR_MAPPING, SccChannel
from ttconv.style_properties import NamedColors, TextDecorationType, \
  FontStyleType, ColorType

_ROW_MAPPING = {
  (0x01, 0x40): 1,
  (0x01, 0x60): 2,
  (0x02, 0x40): 3,
  (0x02, 0x60): 4,
  (0x05, 0x40): 5,
  (0x05, 0x60): 6,
  (0x06, 0x40): 7,
  (0x06, 0x60): 8,
  (0x07, 0x40): 9,
  (0x07, 0x60): 10,
  (0x00, 0x40): 11,
  (0x03, 0x40): 12,
  (0x03, 0x60): 13,
  (0x04, 0x40): 14,
  (0x04, 0x60): 15
}

_ROW_PAC_MAPPING_CH1 = [
  (0x01, 0x40),
  (0x01, 0x60),
  (0x02, 0x40),
  (0x02, 0x60),
  (0x05, 0x40),
  (0x05, 0x60),
  (0x06, 0x40),
  (0x06, 0x60),
  (0x07, 0x40),
  (0x07, 0x60),
  (0x00, 0x40),
  (0x03, 0x40),
  (0x03, 0x60),
  (0x04, 0x40),
  (0x04, 0x60)
]

_PAC_COLOR_MAPPING = {
  NamedColors.white.value: 0,
  NamedColors.green.value: 2,
  NamedColors.blue.value: 4,
  NamedColors.cyan.value: 6,
  NamedColors.red.value: 8,
  NamedColors.yellow.value: 10,
  NamedColors.magenta.value: 12,
}

class _SccPacDescriptionBits:
  """Helper class for SCC PAC description bits handling"""

  def __init__(self, bits: int):
    self._bits = bits

  def get_underline(self) -> bool:
    """Returns whether the PAC description bits sets the underline decoration"""
    return self._bits % 2 == 1

  def get_italic(self) -> bool:
    """Returns whether the PAC description bits sets the italic style"""
    return self._bits in (0x0E, 0x0F)

  def get_color(self) -> Optional[ColorType]:
    """Returns the color from the PAC description bits"""

    if self._bits not in list(range(0x00, 0x10)):
      return None

    if self._bits in (0x00, 0x01, 0x0E, 0x0F):
      return NamedColors.white.value

    return SCC_COLOR_MAPPING.get(self._bits, None)

  def get_indent(self) -> Optional[int]:
    """Returns the column offset from the PAC description bits"""
    if self._bits in list(range(0x10, 0x20)):
      return ((self._bits - 0x10) - (self._bits % 2)) * 2

    return None


class SccPreambleAddressCode:
  """SCC PAC definition"""

  def __init__(self, channel: SccChannel,
               row: int,
               color: ColorType = NamedColors.white,
               indent: Optional[int] = None,
               is_italic: bool = False,
               is_underline: bool = False):
    self._channel = channel
    self._row = row
    self._color = color
    self._indent = indent
    self._is_italic = is_italic
    self._is_underline = is_underline

  def get_row(self) -> int:
    """Returns the PAC row"""
    return self._row

  def get_indent(self) -> Optional[int]:
    """Returns PAC column offset"""
    return self._indent

  def get_color(self) -> Optional[ColorType]:
    """Returns PAC color"""
    return self._color

  def is_underline(self) -> bool:
    """Returns PAC underline flag"""
    return self._is_underline

  def is_italic(self) -> bool:
    """Returns PAC italic flag"""
    return self._is_italic

  def get_font_style(self) -> Optional[FontStyleType]:
    """Returns PAC font style"""
    return FontStyleType.italic if self._is_italic else None

  def get_text_decoration(self) -> Optional[TextDecorationType]:
    """Returns PAC text decoration"""
    return TextDecorationType(underline=True) if self._is_underline else None

  def get_channel(self) -> SccChannel:
    """Returns PAC channel"""
    return self._channel

  def __eq__(self, other) -> bool:
    """Overrides default implementation"""
    return isinstance(other, SccPreambleAddressCode) \
           and self.get_row() == other.get_row() \
           and self.get_indent() == other.get_indent() \
           and self.get_color() == other.get_color() \
           and self.get_font_style() == other.get_font_style() \
           and self.get_text_decoration() == other.get_text_decoration()

  def get_ch1_packet(self):
    hi, lo = _ROW_PAC_MAPPING_CH1[self.get_row() - 1]

    hi = hi + 0x10 # channel 1

    if self.is_italic():
      lo = lo + 0x0E
    elif self.get_indent() is not None:
      lo = lo + 0x10 + self.get_indent() // 2
    else:
      lo = lo + _PAC_COLOR_MAPPING.get(self.get_color(), 0)

    if self.is_underline():
      lo = lo + 1

    return hi * 256 + lo

  @staticmethod
  def find(byte_1: int, byte_2: int) -> Optional[SccPreambleAddressCode]:
    """Find the SCC PAC corresponding to the specified bytes"""

    row = SccPreambleAddressCode._get_row(byte_1, byte_2)
    if row is None:
      # Failed to extract PAC row from specified bytes
      return None

    desc_bits = SccPreambleAddressCode._get_description_bits(byte_2)
    if desc_bits is None:
      # Failed to extract PAC description from specified bytes
      return None

    return SccPreambleAddressCode(
      SccChannel.CHANNEL_2 if byte_1 & 0x08 else SccChannel.CHANNEL_1,
      row,
      desc_bits.get_color(),
      desc_bits.get_indent(),
      desc_bits.get_italic(),
      desc_bits.get_underline()
    )

  @staticmethod
  def _get_row(byte_1: int, byte_2: int) -> Optional[int]:
    """Decodes SCC PAC row number from specified bytes"""
    if byte_1 not in list(range(0x10, 0x20)):
      return None

    row_bits = ((byte_1 & 0x0F) % 0X08, byte_2 & 0x60)

    return _ROW_MAPPING.get(row_bits, None)

  @staticmethod
  def _get_description_bits(byte_2: int) -> Optional[_SccPacDescriptionBits]:
    """Extracts descriptions bits from second byte of the input pair"""
    if byte_2 not in list(range(0x40, 0x80)):
      return None
    return _SccPacDescriptionBits(byte_2 & 0x1F)

  def debug(self, value: int) -> str:
    """Debug representation of the code"""
    debug = "[" + str(self.get_channel()) + "|"
    debug += "PAC|" + str(self.get_row()) + "|" + str(self.get_indent())
    if self.get_color() is not None:
      debug += "|" + str(self.get_color())
    if self.get_font_style() is not None:
      debug += "|I"
    if self.get_text_decoration() is not None:
      debug += "|U"
    debug += "/" + hex(value) + "]"
    return debug



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SCC Control Codes"""

from __future__ import annotations

import typing

from ttconv.scc.codes import SccCode


class SccControlCode(SccCode):
  """SCC Control Code definition"""
  AOF = (0x1422, 0x1C22, 0x1522, 0x1D22)  # Reserved (formerly Alarm Off)
  AON = (0x1423, 0x1C23, 0x1523, 0x1D23)  # Reserved (formerly Alarm On)
  BS = (0x1421, 0x1C21, 0x1521, 0x1D21)  # Backspace
  CR = (0x142D, 0x1C2D, 0x152D, 0x1D2D)  # Carriage Return
  DER = (0x1424, 0x1C24, 0x1524, 0x1D24)  # Delete to End of Row
  EDM = (0x142C, 0x1C2C, 0x152C, 0x1D2C)  # Erase Displayed Memory
  ENM = (0x142E, 0x1C2E, 0x152E, 0x1D2E)  # Erase Non-Displayed Memory
  EOC = (0x142F, 0x1C2F, 0x152F, 0x1D2F)  # End of Caption (Flip Memories)
  FON = (0x1428, 0x1C28, 0x1528, 0x1D28)  # Flash On
  RDC = (0x1429, 0x1C29, 0x1529, 0x1D29)  # Resume Direct Captioning
  RTD = (0x142B, 0x1C2B, 0x152B, 0x1D2B)  # Resume Text Display
  TO1 = (0x1721, 0x1F21, 0x1721, 0x1F21)  # Tab Offset 1 Column
  TO2 = (0x1722, 0x1F22, 0x1722, 0x1F22)  # Tab Offset 2 Columns
  TO3 = (0x1723, 0x1F23, 0x1723, 0x1F23)  # Tab Offset 3 Columns
  TR = (0x142A, 0x1C2A, 0x152A, 0x1D2A)  # Text Restart
  RCL = (0x1420, 0x1C20, 0x1520, 0x1D20)  # Resume caption loading
  RU2 = (0x1425, 0x1C25, 0x1525, 0x1D25)  # Roll-Up Captions-2 Rows
  RU3 = (0x1426, 0x1C26, 0x1526, 0x1D26)  # Roll-Up Captions-3 Rows
  RU4 = (0x1427, 0x1C27, 0x1527, 0x1D27)  # Roll-Up Captions-4 Rows

  def __init__(self, channel_1_field_1: int, channel_2_field_1: int, channel_1_field_2: int, channel_2_field_2: int):
    super().__init__(channel_1_field_1, channel_2_field_1)
    self._channel_1_field_2 = channel_1_field_2
    self._channel_2_field_2 = channel_2_field_2

  def get_name(self) -> str:
    """Retrieves Control Code name"""
    return self.name

  def get_values(self) -> typing.Tuple[int, int, int, int]:
    """Returns SCC Control Code values"""
    return self._channel_1, self._channel_2, self._channel_1_field_2, self._channel_2_field_2

  @staticmethod
  def find(value: int) -> typing.Optional[SccControlCode]:
    """Find the Control Code corresponding to the specified value"""
    for control_code in list(SccControlCode):
      if control_code.contains_value(value):
        return control_code
    return None

  def debug(self, value: int) -> str:
    """Debug representation of the code"""
    return "[" + str(self.get_channel(value)) + "|CC|" + self.get_name() + "/" + hex(value) + "]"



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""IMSC configuration"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from fractions import Fraction
from typing import Optional, Union

from ttconv.config import ModuleConfiguration
from ttconv.imsc.attributes import TimeExpressionSyntaxEnum

class ContentProfilesSignaling(Enum):
  """IMSC profile signaling options"""

  NONE = "none"
  """No content profile is signaled"""

  CONTENT_PROFILES = "content_profiles"
  """Content profiles are signaled using the ttp:contentProfiles attribute"""

  def __init__(self, label: str):
    self.label = label

  @staticmethod
  def from_value(value: Union[str, ContentProfilesSignaling]) -> ContentProfilesSignaling:
    """Return the enum corresponding to the specified value"""
    if isinstance(value, ContentProfilesSignaling):
      return value

    for ps in list(ContentProfilesSignaling):
      if value.lower() == ps.label.lower():
        return ps

    raise ValueError(f"Invalid ContentProfilesSignaling value '{value}'. Expect {','.join([e.value for e in list(ContentProfilesSignaling)])}")

def parse_time_expression_syntax(config_value: str) -> Optional[TimeExpressionSyntaxEnum]:
  """Parse time expression from string value"""
  if config_value is None:
    return config_value

  str_values = map(lambda e: e.value, list(TimeExpressionSyntaxEnum))
  if config_value not in str_values:
    raise ValueError("Invalid time expression format", config_value)

  return TimeExpressionSyntaxEnum[config_value]

@dataclass
class IMSCWriterConfiguration(ModuleConfiguration):
  """IMSC writer configuration"""

  class FractionDecoder:
    """Utility callable for converting string to Fraction"""

    def __call__(self, value: str) -> Optional[Fraction]:
      if value is None:
        return None

      [num, den] = value.split('/')

      return Fraction(int(num), int(den))

  @classmethod
  def name(cls):
    return "imsc_writer"

  time_format: Optional[TimeExpressionSyntaxEnum] = field(
    default=None,
    metadata={"decoder": parse_time_expression_syntax}
    )
  fps: Optional[Fraction] = field(
    default=None,
    metadata={"decoder": FractionDecoder()}
    )
  profile_signaling: ContentProfilesSignaling = field(
    default=ContentProfilesSignaling.NONE,
    metadata={"decoder": ContentProfilesSignaling.from_value}
    )




#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SCC word"""

from __future__ import annotations

from typing import Optional

from ttconv.scc.codes import SccCode, SccChannel
from ttconv.scc.codes.attribute_codes import SccAttributeCode
from ttconv.scc.codes.control_codes import SccControlCode
from ttconv.scc.codes.extended_characters import SccExtendedCharacter
from ttconv.scc.codes.mid_row_codes import SccMidRowCode
from ttconv.scc.codes.preambles_address_codes import SccPreambleAddressCode
from ttconv.scc.codes.special_characters import SccSpecialCharacter
from ttconv.scc.codes.standard_characters import SCC_STANDARD_CHARACTERS_MAPPING

PARITY_BIT_MASK = 0b01111111


class SccWord:
  """SCC hexadecimal word definition"""

  def __init__(self, byte_1: int, byte_2: int):
    self.byte_1 = byte_1
    self.byte_2 = byte_2
    self.value = byte_1 * 0x100 + byte_2
    self.code: Optional[SccCode | SccPreambleAddressCode] = self._find_code()

  @staticmethod
  def _is_hex_word(word: str) -> bool:
    """Checks whether the specified word is a 2-bytes hexadecimal word"""
    if len(word) != 4:
      return False
    try:
      int(word, 16)
    except ValueError:
      return False
    return True

  @staticmethod
  def _decipher_parity_bit(byte: int):
    """Extracts the byte value removing the odd parity bit"""
    return byte & PARITY_BIT_MASK

  @staticmethod
  def from_value(value: int) -> SccWord:
    """Creates a SCC word from the specified integer value"""
    if value > 0xFFFF:
      raise ValueError("Expected a 2-bytes int value, instead got ", hex(value))
    data = value.to_bytes(2, byteorder='big')
    return SccWord.from_bytes(data[0], data[1])

  @staticmethod
  def from_bytes(byte_1: int, byte_2: int) -> SccWord:
    """Creates a SCC word from the specified bytes"""
    if byte_1 > 0xFF or byte_2 > 0xFF:
      raise ValueError(f"Expected two 1-byte int values, instead got {hex(byte_1)} and {hex(byte_2)}")
    byte_1 = SccWord._decipher_parity_bit(byte_1)
    byte_2 = SccWord._decipher_parity_bit(byte_2)

    return SccWord(byte_1, byte_2)

  @staticmethod
  def from_str(hex_word: str) -> SccWord:
    """Extracts hexadecimal word bytes to create a SCC word"""
    if not SccWord._is_hex_word(hex_word):
      raise ValueError("Expected a 2-bytes hexadecimal word, instead got ", hex_word)

    data = bytes.fromhex(hex_word)
    return SccWord.from_bytes(data[0], data[1])

  def _find_code(self) -> Optional[SccCode | SccPreambleAddressCode]:
    """Find corresponding code"""
    if self.is_code():
      return SccControlCode.find(self.value) or \
        SccAttributeCode.find(self.value) or \
        SccMidRowCode.find(self.value) or \
        SccPreambleAddressCode.find(self.byte_1, self.byte_2) or \
        SccSpecialCharacter.find(self.value) or \
        SccExtendedCharacter.find(self.value)
    return None

  def to_text(self) -> str:
    """Converts SCC word to text"""
    return ''.join(SCC_STANDARD_CHARACTERS_MAPPING.get(byte, chr(byte)) for byte in [self.byte_1, self.byte_2] if byte != 0x00)

  def get_code(self) -> Optional[SccCode]:
    """Returns the SCC code if any"""
    return self.code

  def is_code(self) -> bool:
    """Returns true if the word is an SCC code, i.e. the first byte
    is a non-printing character in the range 10h to 1Fh."""
    return 0x10 <= self.byte_1 <= 0x1F

  def get_channel(self) -> Optional[SccChannel]:
    """Returns the caption channel, if the word is an SCC code"""
    if self.code is None:
      return None
    if isinstance(self.code, SccPreambleAddressCode):
      return self.code.get_channel()
    return self.code.get_channel(self.value)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''IMSC style properties'''

import inspect
import math
import typing
import ttconv.imsc.namespaces as xml_ns
import ttconv.utils
import ttconv.style_properties as styles
import ttconv.imsc.utils as utils
import ttconv.model as model

class StyleParsingContext:
  doc: model.ContentDocument

class StyleProperty:
  '''Base class for style properties'''

  @property
  @classmethod
  def model_prop(cls):
    raise NotImplementedError

  @property
  @classmethod
  def ns(cls): # pylint: disable=invalid-name
    raise NotImplementedError

  @property
  @classmethod
  def local_name(cls):
    raise NotImplementedError

  @classmethod
  def has_px(cls, _attrib_value: typing.Any) -> bool:
    return False

  @classmethod
  def extract(cls, context: StyleParsingContext, xml_attrib: str):
    '''Converts an IMSC style property to a data model value'''
    raise NotImplementedError

  @classmethod
  def to_model(cls, context: StyleParsingContext, xml_element) -> typing.Tuple[model.StyleProperty, typing.Any]:
    '''Extracts the value of the style property from a TTML element and returns a tuple consisting of 
    the matching model style property and the value of the model style property'''
    return (
      cls.model_prop,
      cls.extract(context, xml_element.get(f"{{{cls.ns}}}{cls.local_name}"))
    )

  @classmethod
  def from_model(cls, xml_element, model_value):
    '''Converts an data model value to an IMSC style property'''
    raise NotImplementedError

class StyleProperties:
  '''TTML style properties

  Class variables:
  
  `BY_QNAME`: mapping of qualified name to StyleProperty class
  `BY_MODEL_PROP`: mapping of model style property classes to StyleProperty classes
  '''

  class BackgroundColor(StyleProperty):
    '''Corresponds to tts:backgroundColor.'''

    ns = xml_ns.TTS
    local_name = "backgroundColor"
    model_prop = styles.StyleProperties.BackgroundColor

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return ttconv.utils.parse_color(xml_attrib)

    @classmethod
    def from_model(cls, xml_element, model_value: styles.ColorType):
      if model_value != styles.NamedColors.transparent.value:
        xml_element.set(
          f"{{{cls.ns}}}{cls.local_name}",
          StyleProperties.to_ttml_color(model_value)
        )

  class Color(StyleProperty):
    '''Corresponds to tts:color.'''

    ns = xml_ns.TTS
    local_name = "color"
    model_prop = styles.StyleProperties.Color

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return ttconv.utils.parse_color(xml_attrib)

    @classmethod
    def from_model(cls, xml_element, model_value):
      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}",
        StyleProperties.to_ttml_color(model_value)
      )

  class Direction(StyleProperty):
    '''Corresponds to tts:direction.'''

    ns = xml_ns.TTS
    local_name = "direction"
    model_prop = styles.StyleProperties.Direction

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.DirectionType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class Disparity(StyleProperty):
    '''Corresponds to tts:disparity.
    '''

    ns = xml_ns.TTS
    local_name = "disparity"
    model_prop = styles.StyleProperties.Disparity

    @classmethod
    def has_px(cls, attrib_value: styles.LengthType) -> bool:
      return attrib_value.units == styles.LengthType.Units.px 

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return StyleProperties.ttml_length_to_model(context, xml_attrib)

    @classmethod
    def from_model(cls, xml_element, model_value: styles.LengthType):
      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}",
        StyleProperties.to_ttml_length(model_value)
      )
      
      
  class Display(StyleProperty):
    '''Corresponds to tts:display.'''

    ns = xml_ns.TTS
    local_name = "display"
    model_prop = styles.StyleProperties.Display

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.DisplayType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.DisplayType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class DisplayAlign(StyleProperty):
    '''Corresponds to tts:displayAlign.'''

    ns = xml_ns.TTS
    local_name = "displayAlign"
    model_prop = styles.StyleProperties.DisplayAlign

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.DisplayAlignType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.DisplayAlignType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class Extent(StyleProperty):
    '''Corresponds to tts:extent.
    '''

    ns = xml_ns.TTS
    local_name = "extent"
    model_prop = styles.StyleProperties.Extent

    @classmethod
    def has_px(cls, attrib_value: styles.ExtentType) -> bool:
      return attrib_value.height.units == styles.LengthType.Units.px or \
        attrib_value.width.units == styles.LengthType.Units.px

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      if xml_attrib == "auto":

        return styles.ExtentType(
          height=styles.LengthType(1, styles.LengthType.Units.rh),
          width=styles.LengthType(1, styles.LengthType.Units.rw)
        )

      s = xml_attrib.split(" ")

      if len(s) != 2:
        raise ValueError("Bad tts:extent syntax")

      return styles.ExtentType(
        height=StyleProperties.ttml_length_to_model(context, s[1]),
        width=StyleProperties.ttml_length_to_model(context, s[0])
      )

    @classmethod
    def from_model(cls, xml_element, model_value):
      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}", 
        f"{StyleProperties.to_ttml_length(model_value.width)}"
        f" {StyleProperties.to_ttml_length(model_value.height)}"
      )


  class FillLineGap(StyleProperty):
    '''Corresponds to itts:fillLineGap.'''

    ns = xml_ns.ITTS
    local_name = "fillLineGap"
    model_prop = styles.StyleProperties.FillLineGap

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return xml_attrib == "true"

    @classmethod
    def from_model(cls, xml_element, model_value: bool):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", "true" if model_value else "false")


  class FontFamily(StyleProperty):
    '''Corresponds to tts:fontFamily.'''

    ns = xml_ns.TTS
    local_name = "fontFamily"
    model_prop = styles.StyleProperties.FontFamily

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):

      return tuple(
        map(
          lambda f: styles.GenericFontFamilyType.monospaceSerif if f is styles.GenericFontFamilyType.default else f,
          utils.parse_font_families(xml_attrib)
        )
      )

    @classmethod
    def from_model(cls, xml_element, model_value):
      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}",
        utils.serialize_font_family(model_value)
    )


  class FontSize(StyleProperty):
    '''Corresponds to tts:fontSize.'''

    ns = xml_ns.TTS
    local_name = "fontSize"
    model_prop = styles.StyleProperties.FontSize

    @classmethod
    def has_px(cls, attrib_value: styles.LengthType) -> bool:
      return attrib_value.units == styles.LengthType.Units.px 

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return StyleProperties.ttml_length_to_model(context, xml_attrib)

    @classmethod
    def from_model(cls, xml_element, model_value):
      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}", 
        StyleProperties.to_ttml_length(model_value)
      )


  class FontStyle(StyleProperty):
    '''Corresponds to tts:fontStyle.'''

    ns = xml_ns.TTS
    local_name = "fontStyle"
    model_prop = styles.StyleProperties.FontStyle

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.FontStyleType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.FontStyleType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class FontVariant(StyleProperty):
    '''Corresponds to tts:fontVariant.'''

    ns = xml_ns.TTS
    local_name = "fontVariant"
    model_prop = styles.StyleProperties.FontVariant

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.FontVariantType(xml_attrib)

    @classmethod
    def from_model(cls, xml_element, model_value: styles.FontVariantType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class FontWeight(StyleProperty):
    '''Corresponds to tts:fontWeight.'''

    ns = xml_ns.TTS
    local_name = "fontWeight"
    model_prop = styles.StyleProperties.FontWeight

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.FontWeightType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.FontWeightType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class LineHeight(StyleProperty):
    '''tts:lineHeight'''

    ns = xml_ns.TTS
    local_name = "lineHeight"
    model_prop = styles.StyleProperties.LineHeight

    @classmethod
    def has_px(cls, attrib_value: styles.LengthType) -> bool:
      if attrib_value is not styles.SpecialValues.normal:
        return attrib_value.units == styles.LengthType.Units.px

      return False

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.SpecialValues.normal \
        if xml_attrib == "normal" \
        else StyleProperties.ttml_length_to_model(context, xml_attrib)

    @classmethod
    def from_model(cls, xml_element, model_value):
      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}", 
        StyleProperties.to_ttml_length(model_value) \
          if model_value is not styles.SpecialValues.normal \
          else "normal"
      )


  class LinePadding(StyleProperty):
    '''Corresponds to ebutts:linePadding.'''

    ns = xml_ns.EBUTTS
    local_name = "linePadding"
    model_prop = styles.StyleProperties.LinePadding

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      lp = StyleProperties.ttml_length_to_model(context, xml_attrib)

      if lp.units != styles.LengthType.Units.c:
        raise ValueError("ebutts:linePadding must be expressed in 'c'")

      return lp

    @classmethod
    def from_model(cls, xml_element, model_value: styles.LengthType):
      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}",
        StyleProperties.to_ttml_length(model_value)
        )


  class LuminanceGain(StyleProperty):
    '''Corresponds to tts:luminanceGain.'''

    ns = xml_ns.TTS
    local_name = "luminanceGain"
    model_prop = styles.StyleProperties.LuminanceGain

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return float(xml_attrib)

    @classmethod
    def from_model(cls, xml_element, model_value):
      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}",
        f"{model_value}"
      )


  class MultiRowAlign(StyleProperty):
    '''Corresponds to ebutts:multiRowAlign.'''

    ns = xml_ns.EBUTTS
    local_name = "multiRowAlign"
    model_prop = styles.StyleProperties.MultiRowAlign

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.MultiRowAlignType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.MultiRowAlignType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class Opacity(StyleProperty):
    '''Corresponds to tts:opacity.'''

    ns = xml_ns.TTS
    local_name = "opacity"
    model_prop = styles.StyleProperties.Opacity

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return float(xml_attrib)

    @classmethod
    def from_model(cls, xml_element, model_value):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", f"{model_value}")


  class Origin(StyleProperty):
    '''Corresponds to tts:origin.'''

    ns = xml_ns.TTS
    local_name = "origin"
    model_prop = styles.StyleProperties.Origin

    @classmethod
    def has_px(cls, attrib_value: styles.CoordinateType) -> bool:
      return attrib_value.x.units == styles.LengthType.Units.px or \
        attrib_value.y.units == styles.LengthType.Units.px

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):

      if xml_attrib == "auto":

        r = styles.CoordinateType(
          x=styles.LengthType(0, styles.LengthType.Units.pct),
          y=styles.LengthType(0, styles.LengthType.Units.pct)
        )

      else:
        s = xml_attrib.split(" ")

        if len(s) != 2:
          raise ValueError("tts:origin has not two components")

        r = styles.CoordinateType(
          x=StyleProperties.ttml_length_to_model(context, s[0]),
          y=StyleProperties.ttml_length_to_model(context, s[1])
        )

      return r

    @classmethod
    def from_model(cls, xml_element, model_value):
      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}", 
        f"{model_value.x.value:g}{model_value.x.units.value} {model_value.y.value:g}{model_value.y.units.value}"
      )


  class Overflow(StyleProperty):
    '''Corresponds to tts:overflow.'''

    ns = xml_ns.TTS
    local_name = "overflow"
    model_prop = styles.StyleProperties.Overflow

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.OverflowType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.OverflowType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class Padding(StyleProperty):
    '''Corresponds to tts:padding.'''
    
    ns = xml_ns.TTS
    local_name = "padding"
    model_prop = styles.StyleProperties.Padding

    @classmethod
    def has_px(cls, attrib_value: styles.PaddingType) -> bool:
      return attrib_value.after.units == styles.LengthType.Units.px or \
        attrib_value.before.units == styles.LengthType.Units.px or \
        attrib_value.start.units == styles.LengthType.Units.px or \
        attrib_value.end.units == styles.LengthType.Units.px

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      s = xml_attrib.split(" ")

      if len(s) == 1:

        before = end = after = start = StyleProperties.ttml_length_to_model(context, s[0])

      elif len(s) == 2:

        before = after = StyleProperties.ttml_length_to_model(context, s[0])
        end = start = StyleProperties.ttml_length_to_model(context, s[1])

      elif len(s) == 3:

        before = StyleProperties.ttml_length_to_model(context, s[0])
        end = start = StyleProperties.ttml_length_to_model(context, s[1])
        after = StyleProperties.ttml_length_to_model(context, s[2])

      elif len(s) == 4:

        before = StyleProperties.ttml_length_to_model(context, s[0])
        end = StyleProperties.ttml_length_to_model(context, s[1])
        after = StyleProperties.ttml_length_to_model(context, s[2])
        start = StyleProperties.ttml_length_to_model(context, s[3])

      else:

        raise ValueError("Bad syntax")

      return styles.PaddingType(before, end, after, start)

    @classmethod
    def from_model(cls, xml_element, model_value: styles.PaddingType):
      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}",
        f"{StyleProperties.to_ttml_length(model_value.before)}"
        f" {StyleProperties.to_ttml_length(model_value.end)}"
        f" {StyleProperties.to_ttml_length(model_value.after)}"
        f" {StyleProperties.to_ttml_length(model_value.start)}"
      )


  class Position(StyleProperty):
    '''Corresponds to tts:position.'''

    ns = xml_ns.TTS
    local_name = "position"
    model_prop = styles.StyleProperties.Position

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):

      (h_edge, h_offset, v_edge, v_offset) = utils.parse_position(xml_attrib)

      return styles.PositionType(
        h_offset=h_offset,
        v_offset=v_offset,
        h_edge=styles.PositionType.HEdge(h_edge),
        v_edge=styles.PositionType.VEdge(v_edge)
      )

    @classmethod
    def from_model(cls, xml_element, model_value: styles.PositionType):
      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}", 
        f"{model_value.h_edge.value} " \
        f"{model_value.h_offset.value:g}{model_value.h_offset.units.value} " \
        f"{model_value.v_edge.value} " \
        f"{model_value.v_offset.value:g}{model_value.v_offset.units.value}"
      )


  class RubyAlign(StyleProperty):
    '''Corresponds to tts:rubyAlign.'''

    ns = xml_ns.TTS
    local_name = "rubyAlign"
    model_prop = styles.StyleProperties.RubyAlign

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.RubyAlignType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.RubyAlignType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class RubyPosition(StyleProperty):
    '''Corresponds to tts:rubyPosition.'''

    ns = xml_ns.TTS
    local_name = "rubyPosition"
    model_prop = styles.StyleProperties.RubyPosition
  
    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.AnnotationPositionType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.AnnotationPositionType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class RubyReserve(StyleProperty):
    '''Corresponds to tts:rubyReserve.'''

    ns = xml_ns.TTS
    local_name = "rubyReserve"
    model_prop = styles.StyleProperties.RubyReserve

    @classmethod
    def has_px(cls, attrib_value: styles.RubyReserveType) -> bool:
      return attrib_value.length is not None and attrib_value.length.units == styles.LengthType.Units.px

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      
      if xml_attrib == "none":
        return styles.SpecialValues.none

      s = xml_attrib.split(" ")

      if len(s) == 0 or len(s) > 2:
        raise ValueError("Bad syntax")

      rr_pos = styles.RubyReserveType.Position[s[0]]

      rr_length = StyleProperties.ttml_length_to_model(context, s[1]) if len(s) == 2 else None

      return styles.RubyReserveType(rr_pos, rr_length)

    @classmethod
    def from_model(cls, xml_element, model_value):
      if isinstance(model_value, styles.RubyReserveType):
        value = model_value.position.value
        if model_value.length is not None:
          value += f" {StyleProperties.to_ttml_length(model_value.length)}"
      elif model_value == styles.SpecialValues.none:
        value = model_value.value
      else:
        raise TypeError

      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}",
        value
      )


  class Shear(StyleProperty):
    '''Corresponds to tts:shear.'''

    ns = xml_ns.TTS
    local_name = "shear"
    model_prop = styles.StyleProperties.Shear

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):

      (value, units) = utils.parse_length(xml_attrib)

      if units != "%":
        raise ValueError(r"tts:shear must be expressed in % units")

      return value if abs(value) <= 100 else math.copysign(100, value)

    @classmethod
    def from_model(cls, xml_element, model_value: float):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", f"{model_value}%")


  class ShowBackground(StyleProperty):
    '''Corresponds to tts:showBackground.'''

    ns = xml_ns.TTS
    local_name = "showBackground"
    model_prop = styles.StyleProperties.ShowBackground
    
    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.ShowBackgroundType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.VisibilityType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class TextAlign(StyleProperty):
    '''Corresponds to tts:textAlign.'''

    ns = xml_ns.TTS
    local_name = "textAlign"
    model_prop = styles.StyleProperties.TextAlign
  
    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      
      if xml_attrib == "left":

        r = styles.TextAlignType.start

      elif xml_attrib == "right":

        r = styles.TextAlignType.end

      else:
      
        r = styles.TextAlignType[xml_attrib]

      return r

    @classmethod
    def from_model(cls, xml_element, model_value: styles.TextAlignType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class TextCombine(StyleProperty):
    '''Corresponds to tts:textCombine.'''

    ns = xml_ns.TTS
    local_name = "textCombine"
    model_prop = styles.StyleProperties.TextCombine

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.TextCombineType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.TextCombineType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class TextDecoration(StyleProperty):
    '''Corresponds to tts:textDecoration.'''

    ns = xml_ns.TTS
    local_name = "textDecoration"
    model_prop = styles.StyleProperties.TextDecoration




    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      
      if xml_attrib == "none":

        underline = False
        line_through = False
        overline = False

      else:

        s = xml_attrib.split(" ")

        underline = None
        line_through = None
        overline = None

        if "underline" in s:
          underline = True
        elif "noUnderline" in s:
          underline = False
          
        if "lineThrough" in s:
          line_through = True
        elif "noLineThrough" in s:
          line_through = False

        if "overline" in s:
          overline = True
        elif "noOverline" in s:
          overline = False

      return styles.TextDecorationType(
        underline=underline,
        line_through=line_through,
        overline=overline
      )

    @classmethod
    def from_model(cls, xml_element, model_value: styles.TextDecorationType):
      
      actual_values = []

      if model_value.underline is True:
        actual_values.append("underline")
      elif model_value.underline is False:
        actual_values.append("noUnderline")

      if model_value.line_through is True:
        actual_values.append("lineThrough")
      elif model_value.line_through is False:
        actual_values.append("noLineThrough")

      if model_value.overline is True:
        actual_values.append("overline")
      elif model_value.overline is False:
        actual_values.append("noOverline")

      attrib_value = " ".join(actual_values)

      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", attrib_value)


  class TextEmphasis(StyleProperty):
    '''Corresponds to tts:textEmphasis.'''

    ns = xml_ns.TTS
    local_name = "textEmphasis"
    model_prop = styles.StyleProperties.TextEmphasis

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      
      style = None
      style_style = None
      style_symbol = None
      color = None
      position = None

      for c in xml_attrib.split(" "):

        if c == "none":

          return styles.SpecialValues.none

        if c == "auto":

          style = styles.TextEmphasisType.Style.auto

        elif c in ("filled", "open"):
          
          style_style = c

        elif c in ("circle", "dot", "sesame"):

          style_symbol = c

        elif c in styles.TextEmphasisType.Position.__members__:

          position = styles.TextEmphasisType.Position[c]

        elif c == "current":

          color = None

        else:

          color = ttconv.utils.parse_color(c)

      if style_style is None and style_symbol is None:

        style = styles.TextEmphasisType.Style.auto

      else:

        style_style = style_style if style_style is not None else "filled"
        style_symbol = style_symbol if style_symbol is not None else "circle"
        style = styles.TextEmphasisType.Style(f"{style_style} {style_symbol}")

      position = position if position is not None else styles.TextEmphasisType.Position.outside

      return styles.TextEmphasisType(
        style=style,
        color=color,
        position=position
      )

    @classmethod
    def from_model(cls, xml_element, model_value: styles.TextEmphasisType):
      actual_values = []

      actual_values.append(model_value.style.value)

      if model_value.color is not None:
        actual_values.append(StyleProperties.to_ttml_color(model_value.color))

      actual_values.append(model_value.position.value) 

      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}",
        " ".join(actual_values)
      )


  class TextOutline(StyleProperty):
    '''Corresponds to tts:textOutline.'''

    ns = xml_ns.TTS
    local_name = "textOutline"
    model_prop = styles.StyleProperties.TextOutline

    @classmethod
    def has_px(cls, attrib_value: styles.TextOutlineType) -> bool:
      if attrib_value is not styles.SpecialValues.none:
        return attrib_value.thickness.units == styles.LengthType.Units.px

      return False

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str) -> typing.Union[styles.TextOutlineType, styles.SpecialValues]:
      
      if xml_attrib == "none":
        return styles.SpecialValues.none

      s = xml_attrib.split(" ")

      if len(s) == 0 or len(s) > 2:
        raise ValueError("Bad syntax")

      thickness = StyleProperties.ttml_length_to_model(context, s[-1])

      color = ttconv.utils.parse_color(s[0]) if len(s) == 2 else None
      
      return styles.TextOutlineType(
        color=color,
        thickness=thickness
      )

    @classmethod
    def from_model(cls, xml_element, model_value):
      if isinstance(model_value, styles.TextOutlineType):
        values = []

        if model_value.color is not None:
          values.append(StyleProperties.to_ttml_color(model_value.color))

        if model_value.thickness is not None:
          values.append(StyleProperties.to_ttml_length(model_value.thickness))

      elif model_value == styles.SpecialValues.none:
        values = [model_value.value]
      else:
        raise TypeError
      
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", " ".join(values))


  class TextShadow(StyleProperty):
    '''Corresponds to tts:textShadow.'''

    ns = xml_ns.TTS
    local_name = "textShadow"
    model_prop = styles.StyleProperties.TextShadow

    @classmethod
    def has_px(cls, attrib_value: styles.TextShadowType) -> bool:

      for shadow in attrib_value.shadows:
        if shadow.x_offset.units == styles.LengthType.Units.px or \
          shadow.y_offset.units == styles.LengthType.Units.px:
          return True
        if shadow.blur_radius is not None:
          if shadow.blur_radius.units == styles.LengthType.Units.px:
            return True
      
      return False

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str) -> typing.Union[styles.TextShadowType, styles.SpecialValues]:
      
      if xml_attrib == "none":
        return styles.SpecialValues.none

      shadows = []

      for shadow in xml_attrib.split(","):

        cs = shadow.split(" ")

        if len(cs) < 1 or len(cs) > 4:
          raise ValueError("Invalid Syntax")
        
        x_offset = StyleProperties.ttml_length_to_model(context, cs[0])

        y_offset = StyleProperties.ttml_length_to_model(context, cs[1])

        blur_radius = None
        color = None

        if len(cs) == 3:

          try:

            blur_radius = StyleProperties.ttml_length_to_model(context, cs[2])

          except ValueError:

            color = ttconv.utils.parse_color(cs[2])

        else: # len(cs) == 4

          blur_radius = StyleProperties.ttml_length_to_model(context, cs[2])
          color = ttconv.utils.parse_color(cs[3])

        shadows.append(
          styles.TextShadowType.Shadow(
            x_offset=x_offset,
            y_offset=y_offset,
            blur_radius=blur_radius,
            color=color
          )
        )

      return styles.TextShadowType(tuple(shadows))

    @classmethod
    def from_model(cls, xml_element, model_value):
      if isinstance(model_value, styles.TextShadowType):
        actual_values = []

        for shadow in model_value.shadows:
          value = f"{StyleProperties.to_ttml_length(shadow.x_offset)}" \
                    f" {StyleProperties.to_ttml_length(shadow.y_offset)}"
          
          if shadow.blur_radius is not None:
            value += f" {StyleProperties.to_ttml_length(shadow.blur_radius)}"

          if shadow.color is not None:
            value += f" {StyleProperties.to_ttml_color(shadow.color)}"

          actual_values.append(value)

      elif model_value == styles.SpecialValues.none:

        actual_values = [model_value.value]

      else:

        raise TypeError

      xml_element.set(
        f"{{{cls.ns}}}{cls.local_name}",
        ", ".join(actual_values)
      )


  class UnicodeBidi(StyleProperty):
    '''Corresponds to tts:unicodeBidi.'''

    ns = xml_ns.TTS
    local_name = "unicodeBidi"
    model_prop = styles.StyleProperties.UnicodeBidi

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.UnicodeBidiType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.UnicodeBidiType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class Visibility(StyleProperty):
    '''Corresponds to tts:visibility.'''

    ns = xml_ns.TTS
    local_name = "visibility"
    model_prop = styles.StyleProperties.Visibility
    
    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.VisibilityType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.VisibilityType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class WrapOption(StyleProperty):
    '''Corresponds to tts:wrapOption.'''

    ns = xml_ns.TTS
    local_name = "wrapOption"
    model_prop = styles.StyleProperties.WrapOption

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):
      return styles.WrapOptionType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.WrapOptionType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  class WritingMode(StyleProperty):
    '''Corresponds to tts:writingMode.'''

    ns = xml_ns.TTS
    local_name = "writingMode"
    model_prop = styles.StyleProperties.WritingMode

    @classmethod
    def extract(cls, context: StyleParsingContext, xml_attrib: str):

      if xml_attrib == "lr":
        return styles.WritingModeType.lrtb

      if xml_attrib == "rl":
        return styles.WritingModeType.rltb

      if xml_attrib == "tb":
        return styles.WritingModeType.tbrl

      return styles.WritingModeType[xml_attrib]

    @classmethod
    def from_model(cls, xml_element, model_value: styles.WritingModeType):
      xml_element.set(f"{{{cls.ns}}}{cls.local_name}", model_value.value)


  BY_QNAME = {
    f"{{{style_prop.ns}}}{style_prop.local_name}" : style_prop
    for style_name, style_prop in list(locals().items()) if inspect.isclass(style_prop)
    }

  BY_MODEL_PROP = {
    style_prop.model_prop : style_prop
    for style_name, style_prop in list(locals().items()) if inspect.isclass(style_prop) and style_prop.model_prop is not None
    }

  @classmethod
  def ttml_length_to_model(cls, _context: StyleParsingContext, xml_attrib: str):
    (value, units) = utils.parse_length(xml_attrib)

    return styles.LengthType(value, styles.LengthType.Units(units))

  @staticmethod
  def to_ttml_color(model_value: styles.ColorType):
    assert model_value.ident == styles.ColorType.Colorimetry.RGBA8
    color_str = f"#{model_value.components[0]:02x}" \
                 f"{model_value.components[1]:02x}"  \
                 f"{model_value.components[2]:02x}"
    if model_value.components[3] != 0xFF:
      color_str = f"{color_str}{model_value.components[3]:02x}"
    
    return color_str

  @staticmethod
  def to_ttml_length(model_value: styles.LengthType):
    return f"{model_value.value:g}{model_value.units.value}"



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for the app'''

# pylint: disable=R0201,C0115,C0116

import os
import io
import unittest
from contextlib import redirect_stdout
from contextlib import redirect_stderr
import ttconv.tt as tt

class IMSCAppTest(unittest.TestCase):

  def setUp(self):
    if not os.path.exists('build'):
      os.makedirs('build')

  def test_convert(self):
    tt.main(["convert",  
      "-i",  "src/test/resources/ttml/body_only.ttml" ,
      "-o", "build/body_only.out.ttml",
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])

  def test_convert_input_file_type_ttml(self):
    tt.main(["convert",
      "-i", "src/test/resources/ttml/body_only.ttml",
      "--itype", "ttml",
      "-o", "build/body_only.out.ttml",
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])

  def test_convert_input_file_type_TTML(self):
    tt.main(["convert",
      "-i", "src/test/resources/ttml/body_only.ttml",
      "--itype", "TTML",
      "-o", "build/body_only.out.ttml",
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])

  def test_convert_output_file_type_ttml(self):
    tt.main(["convert",
      "-i", "src/test/resources/ttml/body_only.ttml",
      "-o", "build/body_only.out.ttml",
      "--otype", "ttml",
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])

  def test_convert_output_file_type_TTML(self):
    tt.main(["convert",
      "-i", "src/test/resources/ttml/body_only.ttml",
      "-o", "build/body_only.out.ttml",
      "--otype", "TTML",
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])

  def test_convert_input_file_type_scc(self):
    tt.main(["convert", 
      "-i", "src/test/resources/scc/pop-on.scc",
      "--itype", "scc", 
      "-o", "build/pop-on.out.ttml", 
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])

  def test_convert_bad_input_file_name(self):
    with self.assertRaises(ValueError):
      tt.main(["convert", 
      "-i", "src/test/resources/ttml/body_only.not_ttml", 
      "-o", "build/body_only.out.ttml", 
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])

  def test_convert_bad_output_file_name(self):
    with self.assertRaises(ValueError):
      tt.main(["convert", 
        "-i", "src/test/resources/ttml/body_only.ttml", 
        "-o", "build/body_only.out.not_ttml", 
        "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])

  def test_convert_bad_input_file_arg(self):
    with self.assertRaises(ValueError):
      tt.main(["convert", 
        "-i", "src/test/resources/ttml/body_only.ttml", 
        "-o", "build/body_only.out.ttml", 
        "--itype", "not_ttml", 
        "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])

  def test_convert_bad_output_file_arg(self):
    with self.assertRaises(ValueError):
      tt.main(["convert", 
        "-i", "src/test/resources/ttml/body_only.ttml", 
        "-o", "build/body_only.out.not_ttml", 
        "--otype", "not_ttml", 
        "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])

  def test_convert_mismtach_file_type_and_file_name(self):
    tt.main(["convert", 
      "-i", "src/test/resources/ttml/body_only.ttml", 
      "--itype", "scc", 
      "-o", "build/body_only123.out.ttml", 
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])

  def test_pop_on_scc(self):
    tt.main(["convert", 
      "-i", "src/test/resources/scc/pop-on.scc", 
      "-o", "build/pop-on.ttml", 
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])
    self.assertTrue(os.path.exists("build/pop-on.ttml"))
    tt.main(["convert", 
      "-i", "src/test/resources/scc/pop-on.scc", 
      "-o", "build/pop-on.srt", 
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])
    self.assertTrue(os.path.exists("build/pop-on.srt"))

  def test_paint_on_scc(self):
    tt.main(["convert", 
      "-i", "src/test/resources/scc/paint-on.scc", 
      "-o", "build/paint-on.ttml", 
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])
    self.assertTrue(os.path.exists("build/paint-on.ttml"))
    tt.main(["convert", 
      "-i", "src/test/resources/scc/paint-on.scc", 
      "-o", "build/paint-on.srt", 
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])
    self.assertTrue(os.path.exists("build/paint-on.srt"))

  def test_mix_rows_roll_up_scc(self):
    tt.main(["convert", 
      "-i", "src/test/resources/scc/mix-rows-roll-up.scc", 
      "-o", "build/mix-rows-roll-up.ttml", 
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])
    self.assertTrue(os.path.exists("build/mix-rows-roll-up.ttml"))
    tt.main(["convert", 
      "-i", "src/test/resources/scc/mix-rows-roll-up.scc", 
      "-o", "build/mix-rows-roll-up.srt", 
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])
    self.assertTrue(os.path.exists("build/mix-rows-roll-up.srt"))

  def test_bad_function(self):

    stdout = io.StringIO()
    stderr = io.StringIO()
    with redirect_stdout(stdout), redirect_stderr(stderr):
      # Note passing a bad function name
      #
      with self.assertRaises(SystemExit):
        tt.main("covert")

  def test_with_config(self):
    tt.main(['convert', 
      '-i', 'src/test/resources/ttml/body_only.ttml', 
      '-o', 'build/body_only.out.ttml', 
      '--config', '{"general": {"progress_bar":false, "log_level":"WARN"}}'])

  def test_log_levels(self):
    log_level = ["DEBUG", "INFO", "WARN", "ERROR"]
    for l in log_level:
      tt.main(['convert',
        '-i', 'src/test/resources/ttml/body_only.ttml',
        '-o', 'build/body_only.out.ttml',
        '--config', f'{{"general": {{"progress_bar": false, "log_level": "{l}"}} }}'])

  def test_with_config_file(self):
    tt.main(["convert", 
      "-i", "src/test/resources/ttml/body_only.ttml", 
      "-o", "build/body_only.out.ttml", 
      "--config_file", "src/test/resources/config_files/unit_test_cfg.json"])

  def test_file_types_by_type_and_string(self):
    
    self.assertEqual(None, tt.FileTypes.get_file_type(None, None))

    with self.assertRaises(ValueError):
      tt.FileTypes.get_file_type(None, "")

    with self.assertRaises(ValueError):
      tt.FileTypes.get_file_type(None, "asdf")

    self.assertEqual(tt.FileTypes.TTML, tt.FileTypes.get_file_type(tt.FileTypes.TTML.value, None))
    self.assertEqual(tt.FileTypes.TTML, tt.FileTypes.get_file_type(tt.FileTypes.TTML.value, "asdf"))
    self.assertEqual(tt.FileTypes.TTML, tt.FileTypes.get_file_type(None, "ttml"))

    self.assertEqual(tt.FileTypes.SCC, tt.FileTypes.get_file_type(tt.FileTypes.SCC.value, None))
    self.assertEqual(tt.FileTypes.SCC, tt.FileTypes.get_file_type(tt.FileTypes.SCC.value, "asdf"))
    self.assertEqual(tt.FileTypes.SCC, tt.FileTypes.get_file_type(None, "scc"))

    self.assertEqual(tt.FileTypes.SRT, tt.FileTypes.get_file_type(tt.FileTypes.SRT.value, None))
    self.assertEqual(tt.FileTypes.SRT, tt.FileTypes.get_file_type(tt.FileTypes.SRT.value, "asdf"))
    self.assertEqual(tt.FileTypes.SRT, tt.FileTypes.get_file_type(None, "srt"))

  def test_document_lang_override(self):
    out_path = "build/body_only.out.es-419.ttml"
    in_path = "src/test/resources/ttml/body_only.ttml"

    with open(in_path, encoding="utf-8") as f:
      self.assertNotRegex(f.read(), "lang=['\"]es-419['\"]")
    
    tt.main(['convert', 
      '-i', 'src/test/resources/ttml/body_only.ttml', 
      '-o', out_path, 
      '--config', '{"general": {"progress_bar":false, "document_lang":"es-419"}}'])
    
    with open(out_path, encoding="utf-8") as f:
      self.assertRegex(f.read(), "lang=['\"]es-419['\"]")

  def test_lcd_filter(self):
    out_path = "build/referential_styling.ttml"
    in_path = "src/test/resources/ttml/referential_styling.ttml"

    tt.main(['convert',
      '-i', in_path,
      '-o', out_path,
      '--filter', 'lcd',
      '--config', '{"lcd": {"bg_color": "blue", "safe_area": 0, "color": "red", "preserve_text_align": true}}'
      ])

    tt.main(['convert',
      '-i', in_path,
      '-o', out_path,
      '--filter', 'lcd',
      '--config', '{"lcd": {"bg_color":"red"}}'
      ])
    
  def test_srt_reader_extended_tags(self):
    in_path = "src/test/resources/srt/extended-tags.srt"

    out_path = "build/extended_tags.ttml"
    tt.main(['convert',
      '-i', in_path,
      '-o', out_path,
      '--config', '{"srt_reader": {"extended_tags": true}}'
      ])
    with open(out_path, encoding="utf-8") as f:
      self.assertRegex(f.read(), "fontWeight")
    
    out_path = "build/no-extended_tags.ttml"
    tt.main(['convert',
      '-i', in_path,
      '-o', out_path
      ])
    with open(out_path, encoding="utf-8") as f:
      self.assertNotRegex(f.read(), "fontWeight")

  def test_imsc11filter(self):
    out_path = "build/imsc11filter.ttml"
    in_path = "src/test/resources/ttml/referential_styling.ttml"

    tt.main(['convert',
      '-i', in_path,
      '-o', out_path,
      '--filter', 'imsc11filter'
      ])

if __name__ == '__main__':
  unittest.main()
